#include "mpi.h"
#include "mpiimpl.h"

#include <stdlib.h>
#include <string.h>

#define BLOCK_SIZE_PER_NODE (1024*1024) /* allocate 1MB per initial node*/

typedef struct buff_hdr *buff_hdr_ptr;

struct buff_hdr {
  int empty_tag; /* it's realy valid flag */

  struct MPI_msg_hdr msg_hdr;
  char *data;

  buff_hdr_ptr next;  /* equal to NULL only if this is the last entry */
};

/*======================================================================*/

static size_t hdr_size;

static char *rcv_buff;
static size_t rcv_buff_size;

static size_t left;
static size_t left_at_end;
static size_t left_at_beginning;

static buff_hdr_ptr first_entry;
static buff_hdr_ptr last_entry;

/* msg that we last peek at, avoid searching again */
static buff_hdr_ptr last_peek; 

static int compare(buff_hdr_ptr hdr, MPI_Comm comm, int tag, int source);
static void compress(void);

/*======================================================================*/

int MPI_init_rcv_buff(int nodes) {
  hdr_size = sizeof(struct buff_hdr);
  rcv_buff_size = nodes * BLOCK_SIZE_PER_NODE;
  rcv_buff = (char *) malloc(rcv_buff_size);
  if (rcv_buff == NULL) {    
    rcv_buff_size = 0;
#ifdef DEBUG
    printf("Couldn't allocate receive buffer\n");
#endif
    return -1; /* Fail init */
  }
  left_at_end = left = rcv_buff_size;
  left_at_beginning = (size_t)0;
  last_entry = first_entry = NULL;

#ifdef DEBUG
  printf("Allocated bufer of size %d for %d nodes\n", rcv_buff_size, nodes);
#endif
  
  return 0;
}

/*======================================================================*/
/* returns 0 if found -1 if not 
 tag & source are I/O arguments and size is size of found message */
   
int MPI_peek_rcv_buff(MPI_Comm comm, int tag, int source, 
		      struct MPI_msg_hdr *msg_hdr) {
  buff_hdr_ptr tmp;
  
  /* first check in the last peek */
  if ((last_peek != NULL) && compare(last_peek, comm, tag, source)) {
    *msg_hdr = last_peek->msg_hdr;
    return 0;
  }
  
  tmp = first_entry;
  while (tmp != NULL) {
    if (tmp->empty_tag) {
      tmp = tmp->next;
      continue;
    }
    if (compare(tmp,  comm, tag, source)) {
      *msg_hdr = tmp->msg_hdr;
      last_peek = tmp;
      return 0;
    }
    tmp = tmp->next;
  }
  return -1;
}

/*======================================================================*/

int MPI_get_rcv_buff(MPI_Comm comm, int tag, int source, int size, void *buff, 
			struct MPI_msg_hdr *msg_hdr) {
  buff_hdr_ptr tmp;

#ifdef _DEBUG
  compress();
#endif

  /* first check in the last peek */
  if ((last_peek != NULL) && compare(last_peek, comm, tag, source)) {
    if (last_peek->msg_hdr.size > size)
      return MPI_ERR_BUFFER; /* buff to small */
    
    *msg_hdr = last_peek->msg_hdr;
    memcpy(buff, last_peek->data, (size_t)size);
    last_peek->empty_tag = 1; /* tag this is an empty entry */
    last_peek = NULL;
    return MPI_SUCCESS;
  }

  while (tmp != NULL) {
    if (tmp->empty_tag) {
      tmp = tmp->next;
      continue;
    }
    if (compare(tmp,  comm, tag, source)) {
      if (tmp->msg_hdr.size > size) 
	return MPI_ERR_BUFFER; /* buff to small */
      *msg_hdr = tmp->msg_hdr;
      memcpy(buff, tmp->data, (size_t)size);
      tmp->empty_tag = 1; /* tag this as an empty entry */
      return MPI_SUCCESS;
    }
    tmp = tmp->next;
  }
  return -1;
}

/*======================================================================*/
/* add message header to buffer, allocate memory space for message 
   an returns pointer to it.
   returns NULL if not enough space */

void *MPI_put_rcv_buff(struct MPI_msg_hdr *msg_hdr) {
  buff_hdr_ptr tmp;  
  
  size_t entry_size = hdr_size + msg_hdr->size;

  if (left < entry_size) {
#ifdef DEBUG
    printf("Not enough space: left %d, size %d\n", left, entry_size);
#endif
    return NULL; /* not enough space */
  }
  if (left_at_end < entry_size)
    compress(); /* not enough space at the end, make it*/
  
  if (first_entry == NULL) { /* first message */
    tmp = (buff_hdr_ptr)rcv_buff;
    first_entry = tmp;
  }
  else {
    tmp = (buff_hdr_ptr) (last_entry->data + last_entry->msg_hdr.size);
    last_entry->next = tmp;
  }

  /* copy message header */
  tmp->msg_hdr = *msg_hdr;
  tmp->data = (char *)(tmp + hdr_size);
  tmp->next = NULL;
  
  left -= entry_size;
  left_at_end -= entry_size;

  if (last_entry != NULL)
    last_entry->next = tmp;
  last_entry = tmp;
  return tmp->data;
}

/*======================================================================*/
/* move all non-empty messages to the beginning of the buffer */

static void compress(void) {
  buff_hdr_ptr tmp = first_entry ;
  buff_hdr_ptr prev = NULL;
  char *pos = rcv_buff;  
  
  if ((first_entry == NULL) || /* buffer not used yet */
      ((first_entry->empty_tag == 0) && 
       (left == left_at_end))) {  /* all empty space is as the end */
#ifdef _DEBUG
    printf("Compress: Buffer not used yet or all empty space is as the end\n");
#endif
    return; /* nothing to do */ 
  }
  first_entry = NULL; /* reset pointer to first entry */

  while (tmp != NULL) {
    buff_hdr_ptr next = tmp->next;
    size_t entry_size = hdr_size + tmp->msg_hdr.size;

    if (tmp->empty_tag) { /* ignore empty entries */
      tmp = next;
      continue;
    }

#ifdef _DEBUG
    printf("Compress: moves message with tag %d\n", tmp->msg_hdr.tag);
#endif
    memmove((void *)pos, (void *)tmp, entry_size);
    tmp = (buff_hdr_ptr) pos; /* move our tmp ptr to point to the new position */
    tmp->data = pos + hdr_size; /* update pointer to data */

    if (prev != NULL)
      prev->next = tmp; /* update pointer to next of previous entry */
    else 
      first_entry = tmp;/* if there was no prev. this is the first entry */

    last_entry = tmp; /* set the last entry to this one */
    
    pos = pos + entry_size;
    prev = tmp;
    tmp = next;
  }

  /* update globals for buffer */
  left_at_beginning = (size_t) 0;
  left_at_end = left;
  last_peek = NULL; /* invalid */
  return;
}

/*======================================================================*/
/* compare the message in hdr with given argsuments and
   returns 1 if they match, otherwise returns 0 */
   
static int compare(buff_hdr_ptr hdr, MPI_Comm comm, int tag, int source) {

  if (abs(hdr->msg_hdr.comm) == abs(comm)) {
    if (tag == MPI_ANY_TAG) {
      if (source == MPI_ANY_SOURCE)
    return 1;
      return (hdr->msg_hdr.source == source);
    }
   
    if (hdr->msg_hdr.tag == tag) {      
      if (source == MPI_ANY_SOURCE)
    return 1;
      return (hdr->msg_hdr.source == source);
    }
    return 0;
  }
  return 0;
}
