#include "mpi.h"
#include "mpiimpl.h"

/*======================================================================*/

int MPI_Bcast(void *buffer, int count, MPI_Datatype datatype,
	      int root, MPI_Comm comm) {
  int my_rank, d; /* d is dummy */
  int tag = 0; /* msg tag to be used */

  if (count < 0 || datatype < 1 || datatype > MPI_LONG_LONG_INT ||
      comm == MPI_COMM_NULL ||
      !MPI_Comms[comm].intracomm ||
      MPI_Comms[comm].group->nodes[root] == MPI_UNDEFINED) {
    printf("MPI_Bcast(.., %d, %d, %d, %d, %d): invalid argument\n", 
	   count, datatype, root, comm);
    exit(1);
  }

  my_rank = MPI_Comms[comm].group->ranks[MPI_node_num];

  if (my_rank == root) { /* root node for this broadcast */
    int i, group_size = MPI_Comms[comm].group->size;
    size_t data_len = count * MPI_datatype_size[datatype];

    for (i = 0; i < group_size; i++) {
      if (i == root) {
	/* skip me (root)  it is the same buffer */
	continue;
      }
      MPI_internal_snd(comm, i, /*tag*/MPI_BCAST, buffer, data_len, 
		       datatype,/* MPI_BCAST,*/ d, d, d);
        }
    }
  else {
    MPI_Status status;
    MPI_internal_rcv(comm, root, /*tag*/ MPI_BCAST, 
		buffer, count * MPI_datatype_size[datatype], 
		     datatype, /*MPI_BCAST,*/ &d, &d, &d, &status);
  }
  return MPI_SUCCESS;
}

/*======================================================================*/
static int barrier = 0;

int MPI_Barrier (MPI_Comm comm) {
/* send is not synchronized so Bcast wont work anymore 
   but recv still is, so make everybody receive something 
   after we are sure they entered the call */

/* tell scatter & gather this is barrier call */
  barrier = 1; 
  MPI_Gather(NULL, 0, MPI_BYTE, NULL, 0, MPI_BYTE, 0, comm);

/* everybody have entered the call, by sending dummy gather, 
   now tell them they can continue 			       */
  MPI_Scatter(NULL, 0, MPI_BYTE, NULL, 0, MPI_BYTE, 0, comm);
  barrier = 0;

  return MPI_SUCCESS;
}

/*======================================================================*/

int MPI_Gather(void *sendbuf, int sendcount, MPI_Datatype sendtype,
	       void *recvbuf, int recvcount, MPI_Datatype recvtype, 
	       int root, MPI_Comm comm) {

  int my_rank, d, tag = 0;
  int type = barrier ? MPI_BARRIER : MPI_GATHER;
    
  if (comm == MPI_COMM_NULL ||
      !MPI_Comms[comm].intracomm ||
      sendtype < 1 || sendtype > MPI_LONG_LONG_INT ||
      MPI_Comms[comm].group->nodes[root] == MPI_UNDEFINED) {
    printf("MPI_Gather(.., %d, %d,..,%d, %d,%d, %d): invalid argument\n", 
	   sendcount, sendtype, recvcount, recvtype, root, comm);
    exit(1);
  }
  
  my_rank = MPI_Comms[comm].group->ranks[MPI_node_num];

  if (my_rank == root) { /* root, gatherer */
    int i, group_size = MPI_Comms[comm].group->size;
    char *tmp;
    size_t data_len = recvcount * MPI_datatype_size[recvtype];
    MPI_Status status;
        
    if (sendtype != recvtype) {
      printf("MPI_Gather(.., %d, %d,..,%d, %d,%d, %d): invalid argument\n", 
	     sendcount, sendtype, recvcount, recvtype, root, comm);
      exit(1);
    }

    tmp = (char *)recvbuf;
    for (i = 0; i < group_size; i++) {
      /* move to next place in buffer*/
      if (i == root) /* copy data from itself */
	memcpy(tmp, sendbuf, data_len);
      else      
	MPI_internal_rcv(comm, i, /*tag*/type, tmp, data_len,
			 recvtype, /*type,*/ &d, &d, &d, &status);
      tmp += data_len;
    }
  }
  else { /* sender */
    MPI_internal_snd(comm, root, /*tag*/type, sendbuf, 
		     sendcount * MPI_datatype_size[sendtype],
		     sendtype, /*type,*/ d, d, d);
  }
  return MPI_SUCCESS;
}
/*======================================================================*/

int MPI_Gatherv(void *sendbuf, int sendcount, MPI_Datatype sendtype,
		void *recvbuf, int *recvcounts, int *displs,
		MPI_Datatype recvtype, int root, MPI_Comm comm) {

  int my_rank, d, tag = 0;
    
    if (comm == MPI_COMM_NULL ||
        !MPI_Comms[comm].intracomm ||
        sendtype < 1 || sendtype > MPI_LONG_LONG_INT ||
        MPI_Comms[comm].group->nodes[root] == MPI_UNDEFINED) {
      printf("MPI_Gatherv(.., %d, %d,..,..,..,%d, %d): invalid argument\n", 
	       sendcount, sendtype, recvtype, root, comm);
      exit(1);
    }

    my_rank = MPI_Comms[comm].group->ranks[MPI_node_num];
    
    if (my_rank == root) {
      int i, group_size = MPI_Comms[comm].group->size;
      size_t data_len, type_len = MPI_datatype_size[recvtype];
      MPI_Status status;
      void *tmp;
      
      if (sendtype != recvtype) {
	printf("MPI_Gatherv(.., %d, %d,..,..,..,%d, %d): invalid argument\n", 
	       sendcount, sendtype, recvtype, root, comm);
	exit(1);
      }
      for (i = 0; i < group_size; i++) {
	data_len = recvcounts[i] * type_len;
	tmp = ((char *)recvbuf) + displs[i];
	if (i == root)
	  memcpy(tmp, sendbuf, data_len);
	else            
	  MPI_internal_rcv(comm, i, /*tag*/ MPI_GATHERV, tmp, data_len,
			   recvtype, /*MPI_GATHERV,*/ 
			&d, &d, &d, &status);
        }
    }
    else {
      MPI_internal_snd(comm, root, /*tag*/MPI_GATHERV, sendbuf, 
			sendcount * MPI_datatype_size[sendtype],
		       sendtype, /*MPI_GATHER,*/ d, d, d);
    }
    return MPI_SUCCESS;
}

/*======================================================================*/

int MPI_Scatter(void *sendbuf, int sendcount, MPI_Datatype sendtype,
		void *recvbuf, int recvcount, MPI_Datatype recvtype, 
		int root, MPI_Comm comm) {
    int my_rank, d, tag = 0;    
    int type = barrier ? MPI_BARRIER : MPI_GATHER;
    
    if (comm == MPI_COMM_NULL ||
        !MPI_Comms[comm].intracomm ||
        recvtype < 1 || recvtype > MPI_LONG_LONG_INT ||
        MPI_Comms[comm].group->nodes[root] == MPI_UNDEFINED) {
      printf("MPI_Scatter(.., %d, %d,..,%d, %d,%d, %d): invalid argument\n", 
	     sendcount, sendtype, recvcount, recvtype, root, comm);
      exit(1);
    }

    my_rank = MPI_Comms[comm].group->ranks[MPI_node_num];

    if (my_rank == root) {
      int i, group_size = MPI_Comms[comm].group->size;
      char *tmp;
      size_t data_len = sendcount * MPI_datatype_size[sendtype];
        
      if (sendtype != recvtype || sendcount != recvcount) {
	printf("MPI_Scatter(.., %d, %d,..,%d, %d,%d, %d): invalid argument\n", 
	       sendcount, sendtype, recvcount, recvtype, root, comm);
	exit(1);
      }
      
      tmp = (char *)sendbuf;
      for (i = 0; i < group_size; i++) {	
	if (i == root)
	  memcpy(recvbuf, tmp, data_len);
	else     
	  MPI_internal_snd(comm, i, /*tag*/type, tmp, data_len,
			   sendtype, /*type,*/ d, d, d);
	tmp += data_len;
      }
    }
    else {
      MPI_Status status;
      MPI_internal_rcv(comm, root, /*tag*/type, recvbuf, 
		       recvcount * MPI_datatype_size[recvtype],
		       recvtype, /*type,*/ &d, &d, &d, &status);;
    }
    return MPI_SUCCESS;
}

/*======================================================================*/

int MPI_Scatterv(void *sendbuf, int *sendcounts, int *displs,
		 MPI_Datatype sendtype, void *recvbuf, int recvcount,
		 MPI_Datatype recvtype, int root, MPI_Comm comm) {
    int my_rank, d, tag = 0;
    
    if (comm == MPI_COMM_NULL ||
        !MPI_Comms[comm].intracomm ||
        recvtype < 1 || recvtype > MPI_LONG_LONG_INT ||
        MPI_Comms[comm].group->nodes[root] == MPI_UNDEFINED) {
      printf("MPI_Scatterv(..,..,.., $d,.., $d,%d, %d): invalid argument\n", 
	     sendtype, recvcount, recvtype, root, comm);
      exit(1);
    }
    
    my_rank = MPI_Comms[comm].group->ranks[MPI_node_num];
    
    if (my_rank == root) {
      int i, group_size = MPI_Comms[comm].group->size;
      size_t data_len, type_len = MPI_datatype_size[sendtype];
      void *tmp;
        
      if (sendtype != recvtype) {
	printf("MPI_Scatterv(..,..,.., $d,.., $d,%d, %d): invalid argument\n", 
	       sendtype, recvcount, recvtype, root, comm);
	exit(1);
      }
      for (i = 0; i < group_size; i++) {
	data_len = sendcounts[i] * type_len;
	tmp = ((char *)sendbuf) + displs[i];
	if (i == my_rank)
	  memcpy(recvbuf, tmp, data_len);
	else            
	  MPI_internal_snd(comm, i, /*tag*/MPI_SCATTERV, tmp, data_len,
			   sendtype, /*MPI_SCATTERV,*/ d, d, d);
      }
    }
    else {
      MPI_Status status;
      MPI_internal_rcv(comm, root, /*tag*/MPI_SCATTERV, recvbuf, 
			recvcount * MPI_datatype_size[sendtype],
		       recvtype, /*MPI_SCATTERV,*/ &d, &d, &d, &status);
    }
    return MPI_SUCCESS;
}

/*======================================================================*/
