#include "mpi.h"
#include "mpiimpl.h"

struct MPI_Comm_entry MPI_Comms[MPI_MAX_COMMS];

MPI_Comm MPI_first_available_comm = 0;

/*========================================================================================*/

void init_comms() {
  int i;
  for (i = 0; i < MPI_MAX_COMMS; i++)
    MPI_Comms[i].used = 0;
  
  MPI_first_available_comm = 0;
}

/*========================================================================================*/

int MPI_Comm_size(MPI_Comm comm, int *size) {
  if (comm == MPI_COMM_NULL) {
    *size = MPI_UNDEFINED;
    return MPI_ERR_COMM;
  }
  *size = MPI_Comms[comm].local->size;
  return MPI_SUCCESS;
}

/*========================================================================================*/

int MPI_Comm_rank(MPI_Comm comm, int *rank) {
  if (comm == MPI_COMM_NULL) {
    *rank = MPI_UNDEFINED;
    return MPI_ERR_COMM;
  }
  *rank = MPI_Comms[comm].source;
  return MPI_SUCCESS;
}

/*========================================================================================*/

int MPI_Comm_group(MPI_Comm comm, MPI_Group *group) {
  if (comm == MPI_COMM_NULL) {
    *group = MPI_GROUP_NULL;
    return MPI_ERR_COMM;
  }
  *group = MPI_Comms[comm].local;
  (*group)->references++;
  return MPI_SUCCESS;
}

/*========================================================================================*/

int MPI_Comm_dup(MPI_Comm comm, MPI_Comm *new) {
  struct MPI_Comm_entry entry, *new_entry;
  MPI_Status status;
  int d, tag = 0;

  if (comm == MPI_COMM_NULL) {
    printf("Error: duplicating non communicator\n");
    exit(1);
  }
  if (MPI_first_available_comm >= MPI_MAX_COMMS) {
    printf("Error: Can not allocate more communicators\n");
    exit(1);
  }

  entry = MPI_Comms[comm];

  /* do sinhronization to find comm's first_available entry */  
  if (entry.source == 0) { /* root */
    int i;
    MPI_Comm available;
    int group_size = entry.local->size;

    /* found out max of first available in the group */
    for (i = 1; i < group_size; i++) {
      MPI_internal_rcv(comm, MPI_ANY_SOURCE, MPI_COMM_DUP, NULL, 0, MPI_BYTE,
		       &available, &d, &d, &status);
      if (available > MPI_first_available_comm)
	MPI_first_available_comm = available;
    }

    /* send that max to the other nodes in group */
    for (i = 1; i < group_size; i++) 
      MPI_internal_snd(comm, i, MPI_COMM_DUP, NULL, 0, MPI_BYTE,
		       MPI_first_available_comm, d, d);
  }
  else { /* not root of the group */

    /* send to root my first_available comm entry */
    MPI_internal_snd(comm, 0, MPI_COMM_DUP, NULL, 0, MPI_BYTE,
		     MPI_first_available_comm, d, d);

    /* receive from root group's first_available comm entry */
    MPI_internal_rcv(comm, 0, MPI_COMM_DUP, NULL, 0, MPI_BYTE,
              	&MPI_first_available_comm, &d, &d, &status);     
  }

  *new = MPI_first_available_comm++;
  MPI_first_available_comm;
  new_entry = &MPI_Comms[*new];
  
  /* now copy */
  new_entry->group = entry.group;
  new_entry->group->references++;
  
  new_entry->local = entry.local;
  new_entry->local->references++;
  
  new_entry->rcv_context = *new;
  
  if (entry.intracomm) 
    new_entry->snd_context = *new;    
  else /* MPI_COMM of remote group stay the same */ 
    new_entry->snd_context = entry.snd_context;    

  new_entry->source = entry.source;
  new_entry->intracomm = entry.intracomm;
  new_entry->used = entry.used;
  new_entry->tag = entry.tag;
  return MPI_SUCCESS;
}
/*========================================================================================*/

int MPI_Comm_create(MPI_Comm comm, MPI_Group group, MPI_Comm *new) {
  int my_rank, d, tag = 0;
  MPI_Status status;
  struct MPI_Comm_entry entry, *new_entry;

  if (comm == MPI_COMM_NULL) {
    printf("Error: creating communicator from non communicator\n");
    exit(1);
  }
  if (group == MPI_GROUP_NULL) {
    printf("Error: creating communicator from non group\n");
    exit(1);
  }  
  if (MPI_first_available_comm >= MPI_MAX_COMMS) {
    printf("Can not allocate more communicators\n");
    exit(1);
  }

  entry = MPI_Comms[comm];
  if (!entry.intracomm) { /* comm has to be intracomm */
    printf("Error: creating communicator using intercommunicator \n");
    exit(1);
  }

  my_rank = group->ranks[MPI_node_num];
  
  /* do sinhronization to find group's first_available */  
  if (entry.source == 0) { /* root */
    MPI_Comm available;
    int i, group_size = entry.group->size;

    /* found out max of first available in the group */
    for (i = 1; i < group_size; i++) {
      MPI_internal_rcv(comm, i, MPI_COMM_CREATE, NULL, 0, MPI_BYTE,
		       &available, &d, &d, &status);
      if ((available != MPI_UNDEFINED) && 
	  (available > MPI_first_available_comm))
	MPI_first_available_comm = available;
    }
    /* send that max to the other nodes on group */
    for (i = 1; i < group_size; i++) 
      MPI_internal_snd(comm, i, MPI_COMM_CREATE, NULL, 0, MPI_BYTE,
		        MPI_first_available_comm, d, d);
  }
  else {
    MPI_Comm available = (my_rank == MPI_UNDEFINED) ? /* not a group member */
      MPI_UNDEFINED : MPI_first_available_comm;

    /* send to root my first_available comm entry */
    MPI_internal_snd(comm, 0, MPI_COMM_CREATE, NULL, 0, MPI_BYTE,
		     available, d, d);

    /* receive from root group's first_available comm entry */
    MPI_internal_rcv(comm, 0, MPI_COMM_CREATE, NULL, 0, MPI_BYTE,
		     (my_rank == MPI_UNDEFINED) ? &d : &MPI_first_available_comm,
		     &d, &d, &status);

  }

  if (my_rank == MPI_UNDEFINED) { /* for non-members finish */
    *new = MPI_COMM_NULL;
    return MPI_SUCCESS;
  }

  *new = MPI_first_available_comm;
  MPI_first_available_comm++;
  new_entry = &MPI_Comms[*new];
  
  new_entry->group = group;
  new_entry->group->references++;
  
  new_entry->local = group;
  new_entry->local->references++;
  
  new_entry->rcv_context = *new;
  new_entry->snd_context = *new;    

  new_entry->source = my_rank;
  new_entry->intracomm = 1;
  new_entry->used = entry.used;
  new_entry->tag = entry.tag;
  
  return MPI_SUCCESS;
}

/*========================================================================================*/

int MPI_Comm_free(MPI_Comm *comm) {
  struct MPI_Comm_entry entry;

  if (*comm == MPI_COMM_NULL)
    return MPI_ERR_COMM;

  entry = MPI_Comms[*comm];
  MPI_Group_free(&entry.group);
  MPI_Group_free(&entry.local);
  entry.used = 0;
  
  while (!MPI_Comms[MPI_first_available_comm - 1].used)
    MPI_first_available_comm--;

  *comm = MPI_COMM_NULL;
  return MPI_SUCCESS;
}
