#include "mpi.h"
#include "mpiimpl.h"

/*========================================================================================*/

/* this implementation will ignore ordering by keys */
int MPI_Comm_split(MPI_Comm comm, int color, int key, MPI_Comm *newcomm) {
  char tmp[MPI_PACK_GROUP_SIZE];

  struct MPI_Comm_entry entry, *new_entry;
  MPI_Group group;
  int my_rank, d, tag = 0;
  MPI_Status status;

  if (comm == MPI_COMM_NULL) {
    printf("Error: spliting non communicator\n");
    exit(1);
  }
  if (MPI_first_available_comm == MPI_MAX_COMMS) {
    printf("Can not allocate more communicators\n");
    exit(1);
  }
  entry = MPI_Comms[comm];
  if (!entry.intracomm) { /* can split only intracomms */
    printf("Error: spliting intercommunicator\n");
    exit(1);
  }  

  if (entry.source == 0) { /* root */
    MPI_Comm available;
    MPI_Group curr_group;
    MPI__INT colors[MPI_MAX_NODES], curr_color;
    MPI__INT keys[MPI_MAX_NODES], curr_key;
    int curr_rank, curr_node;

    int i, group_size = entry.group->size;

    for (i = 0; i < MPI_MAX_NODES; i++) {
      colors[i] = MPI_UNDEFINED;
      keys[i] = MPI_UNDEFINED;
    }
        
    /* find out max of first available comm entries and collect colors and keys */
    for (i = 1; i < group_size; i++) {
      MPI_internal_rcv(comm, i, MPI_COMM_SPLIT, NULL, 0, MPI_BYTE,
               		&available, &curr_color, &curr_key, &status);
      if (available > MPI_first_available_comm)
        MPI_first_available_comm = available;
      curr_node = entry.group->nodes[status.MPI_SOURCE];
      colors[curr_node] = curr_color;
      keys[curr_node] = curr_key;
    }    
    colors[MPI_node_num] = color; /* at the end add my color */
    keys[MPI_node_num] = key;     /* and key              */
    
    /* now make group for every color and send it to nodes in that group */
    i = 0;
    while (1) {
      int j;
      curr_color = MPI_UNDEFINED;

      while ((curr_color == MPI_UNDEFINED) &&  /* find first colored entry */
         (i < MPI_num_nodes))
        curr_color = colors[i++];

      if (curr_color == MPI_UNDEFINED) {  /* already found every color */
        break;
      }
      curr_group = NULL;
      curr_group = MPI_getGroup();

      /* this implementation will ignore ordering by keys */
      curr_rank = 0;
      for (j = i - 1; (j < MPI_num_nodes) && (curr_rank < group_size);j++) {
        if (colors[j] == curr_color) { /* find all same-colored nodes */
          curr_group->ranks[j] = curr_rank; /* and put them into the group     */
          curr_group->nodes[curr_rank] = j;
          colors[j] = MPI_UNDEFINED;   /* clear color of that node   */
          curr_rank++;
        }
      }
      curr_group->size = curr_rank;
      
      /* pack group */
      MPI_pack_group(curr_group, &tmp[0]);

      /* send it to all members of group */
      for (j = 0; j < curr_group->size; j++) {
        curr_node = curr_group->nodes[j];
        if (curr_node == MPI_node_num) {/* avoid sending to myself, but remember it*/ 
          group = curr_group;
          (group->references)++;
          my_rank = 0; /* stay root in a new group */
          continue;
	}
        /* get the node rank in the comm's context */
        curr_rank = entry.group->ranks[curr_node];

        MPI_internal_snd(comm, curr_rank, MPI_SEND_GROUP, 
		(void *)&tmp[0], MPI_PACK_GROUP_SIZE, MPI_BYTE, 
		MPI_first_available_comm, j, d);
      }
      available++; /* put all other comms after max available comm index */
      MPI_Group_free(&curr_group);
    }
  }
  else { /* not root */
    MPI_Comm available;
        
    /* do not put this node in any group if color == MPI_UNDEFINED */
    available = (color == MPI_UNDEFINED) ?  MPI_COMM_NULL : MPI_first_available_comm;

    /* snd to root my first_available comm entry, color and key */
    MPI_internal_snd(comm, 0, MPI_COMM_SPLIT, NULL, 0, 
             MPI_BYTE, available, color, key);

    if (color == MPI_UNDEFINED) { 
      *newcomm = MPI_COMM_NULL; /* this node is done */
      return MPI_SUCCESS;
    }
    /* receive from root first_available comm entry, comm's group  and rank in it */
    MPI_internal_rcv(comm, 0, MPI_SEND_GROUP, 
		(void *)&tmp[0], MPI_PACK_GROUP_SIZE, MPI_BYTE,
		&MPI_first_available_comm, &my_rank, &d, &status);

    group = MPI_getGroup();
    MPI_unpack_group(&tmp[0], &group);
  }

  if (MPI_first_available_comm >= MPI_MAX_COMMS) {
    printf("Can not allocate more communicators\n");
    exit(1);
  }

  *newcomm = MPI_first_available_comm++;
  new_entry = &MPI_Comms[*newcomm];
  
  new_entry->group = group;
  (new_entry->group->references)++;
  
  new_entry->local = group;
  (new_entry->local->references)++;

  MPI_Group_free(&group);
  
  new_entry->rcv_context = *newcomm;
  new_entry->snd_context = *newcomm;    

  new_entry->source = my_rank;
  new_entry->intracomm = 1;
  new_entry->used = 1;
  new_entry->tag = entry.tag;
  
  return MPI_SUCCESS;  
}

/*========================================================================================*/

int MPI_Comm_test_inter(MPI_Comm comm, int *flag) {
  if (comm == MPI_COMM_NULL) {
    *flag = MPI_UNDEFINED;
    return MPI_ERR_COMM;
  }
  *flag = !MPI_Comms[comm].intracomm;
  return MPI_SUCCESS;
}

/*========================================================================================*/

int MPI_Comm_remote_size(MPI_Comm comm, int *size) {
  if ((comm == MPI_COMM_NULL) ||
      (MPI_Comms[comm].intracomm)){
    *size = MPI_UNDEFINED;
    return MPI_ERR_COMM;
  }
  *size = MPI_Comms[comm].group->size;
  return MPI_SUCCESS;
}

/*========================================================================================*/

int MPI_Comm_remote_group(MPI_Comm comm, MPI_Group *group){
  if ((comm == MPI_COMM_NULL) ||
      (MPI_Comms[comm].intracomm)){
    *group = MPI_GROUP_NULL;
    return MPI_ERR_COMM;
  }
  
  *group = MPI_Comms[comm].group;
  (*group)->references++;
  return MPI_SUCCESS;
}

/*========================================================================================*/
