
#include "mpi.h"
#include "mpiimpl.h"

#define DEBUG

#include <netinet/in.h>

static void sendGroup(MPI_Comm peer_comm, int remote_leader, 
			MPI_Group local_group);

static void recvGroup(MPI_Comm peer_comm, int remote_leader, 
		MPI_Group remote_group, MPI_Comm *remote_comm);


/*========================================================================================*/

int MPI_Intercomm_create(MPI_Comm local_comm, int local_leader, 
             MPI_Comm peer_comm, int remote_leader, int tag,
             MPI_Comm *newintercomm) {
  int i, j, d;
  struct MPI_Comm_entry local_entry, peer_entry, *new_entry;
  MPI_Comm remote_comm;
  MPI_Group remote_group = MPI_getGroup();
  int high;
  int remote_group_size;
  MPI_Status status;
 
  if (MPI_first_available_comm >= MPI_MAX_COMMS) {
    printf("Can not allocate more communicators\n");
    exit(1);
  }
  if ((local_comm == MPI_COMM_NULL) ||
      (peer_comm == MPI_COMM_NULL)) {
    printf("Error: creating intercommunicator from non communicator\n");
    exit(1);
  }
  if ((local_leader == MPI_ANY_SOURCE) || /* no wild cards allowed */
      (remote_leader == MPI_ANY_SOURCE) ||
      (tag == MPI_ANY_TAG)) { 
    printf("Error: creating intercommunicator using wild-cards\n");
    exit(1);
  }
    
  local_entry = MPI_Comms[local_comm];
  peer_entry = MPI_Comms[peer_comm];
  if (!peer_entry.intracomm || /* both comms has to be intracomms */
      !local_entry.intracomm) {
    printf("Error: creating intercommunicator from non intracommunicators\n");
    exit(1);
  }
   
  if (local_entry.source == local_leader) { /* this is a local leader */
    int i;
    MPI_Comm available;
    int local_group_size = local_entry.local->size;
    /* leaders rank in peer comm */
    int leader_peer_rank = peer_entry.group->ranks[MPI_node_num]; 

    if ((leader_peer_rank == MPI_UNDEFINED) ||
        (peer_entry.group->nodes[remote_leader] == MPI_UNDEFINED)) {
        printf("Error: both remote leader and local leader "
                "has to belong to peer communicator\n");
        exit(1);
    }
#ifdef DEBUG
    printf("Leader of %d started intercomm_create\n", local_comm);
#endif    
    /* found out max of first available in the local group */
    for (i = 0; i < local_group_size; i++) {
        if (i == local_leader)
            continue;
        MPI_internal_rcv(local_comm, i, MPI_INTERCOMM_CREATE, NULL, 0, MPI_BYTE,
             		&available, &d, &d, &status);
        if (available > MPI_first_available_comm)
            MPI_first_available_comm = available;
    }
#ifdef DEBUG
     printf("Leader of comm %d agread on new comm entry %d\n", local_comm, MPI_first_available_comm);
     printf("Peer rank %d, leader rank %d\n", leader_peer_rank, remote_leader);
#endif
    /* receive/send remote comm index and remote group from peer leader */
    if (leader_peer_rank > remote_leader) { /* to avoid deadlock */
	high = 0;
	sendGroup(peer_comm, remote_leader, local_entry.group);
	recvGroup(peer_comm, remote_leader, remote_group, &remote_comm);
	MPI_propagateGroup(local_comm, local_leader, remote_group, remote_comm, high); 
    }
    else {
	high = 1;
	recvGroup(peer_comm, remote_leader, remote_group, &remote_comm);
	MPI_propagateGroup(local_comm, local_leader, remote_group,remote_comm, high); 
	sendGroup(peer_comm, remote_leader, local_entry.group);
    }
#ifdef DEBUG
    printf("Leader of comm %d finished intercomm_create\n", local_comm);
#endif
  }
  else { /* followers; source =/= local_leader */    
    size_t addr_size = sizeof(struct sockaddr_in);
#ifdef DEBUG
    printf("Follower of %d started intercomm_create\n", local_comm);
#endif    
    /* send to local leader my first_available comm entry */
    MPI_internal_snd(local_comm, local_leader, MPI_INTERCOMM_CREATE, 
		NULL, 0, MPI_BYTE, MPI_first_available_comm, d, d);

    MPI_internal_rcv(local_comm, local_leader, MPI_INTERCOMM_CREATE, NULL,
		0, MPI_BYTE, &MPI_first_available_comm, &high, &d, &status);

    /* receive from leader max available comm entry,
       remote comm entry, and remote group */
    MPI_internal_rcv(local_comm, local_leader, MPI_INTERCOMM_CREATE,
		NULL, 0, MPI_BYTE, 
		&remote_comm, &(remote_group->size), &d, &status);
#ifdef DEBUG
     printf("Follower of comm %d agread on new comm entry %d\n", local_comm, MPI_first_available_comm);
#endif    

     for (i = 0; i < remote_group->size; i++) {
	/* put new remote group at the end */
	MPI_internal_rcv(local_comm, local_leader, MPI_INTERCOMM_CREATE,	
	   (char *)&MPI_nodes[MPI_num_nodes].listener_addr, addr_size,
	   MPI_BYTE, &d, &d, &d, &status);
        /* find if we already know about the address */
	for (j = 0; j < MPI_num_nodes; j++)
	   if (!memcmp(&MPI_nodes[j].listener_addr,
		       &MPI_nodes[MPI_num_nodes].listener_addr, addr_size))
	      break;
	if (j == MPI_num_nodes)
	    MPI_num_nodes++; /* new (unknown) node */
	remote_group->nodes[i] = j;
	remote_group->ranks[j] = i;
    }
#ifdef DEBUG
    printf("Follower of comm %d finished intercomm_create\n", local_comm);
#endif
  }

  *newintercomm = MPI_first_available_comm++;
  new_entry = &MPI_Comms[*newintercomm];
  
  new_entry->group = remote_group;
  new_entry->group->references++;
  MPI_Group_free(&remote_group);
  
  new_entry->local = local_entry.group;
  new_entry->local->references++;
  
  new_entry->rcv_context = *newintercomm;
  new_entry->snd_context = remote_comm;

  new_entry->source = local_entry.source; /* keep the rank */
  new_entry->intracomm = 0;
  new_entry->used = 1;
  new_entry->tag = high ? MPI_HIGH_TAG : 0;
  
  return MPI_SUCCESS;
}

/*========================================================================*/
/* send listener addresses of members of the group */

static void sendGroup(MPI_Comm comm, int rank, MPI_Group group) {
   int i, d, node;
   int size = group->size;
   size_t addr_size = sizeof(struct sockaddr_in);

   MPI_internal_snd(comm, rank, MPI_INTERCOMM_CREATE, NULL, 0, MPI_BYTE,
		MPI_first_available_comm, size, d);

   for (i = 0; i < size; i++) {
     node = group->nodes[i];
     MPI_internal_snd(comm, rank, MPI_INTERCOMM_CREATE, 
	(char *)&MPI_nodes[node].listener_addr, addr_size, MPI_BYTE,
	d, d, d);
   }
#ifdef DEBUG
   printf("In comm %d group sent to %d\n", comm, rank);
#endif
}

/*========================================================================*/
/* First sends to  local group new local comm, remote comm & remote group
   size. Then get remote group listener addresses and sends them to
   every member of local group */

static void recvGroup(MPI_Comm peer_comm, int remote_leader,
		      MPI_Group remote_group, MPI_Comm *remote_comm) {

    int i, j, d;
    MPI_Status status;
    size_t addr_size = sizeof(struct sockaddr_in);

    MPI_internal_rcv(peer_comm, remote_leader,
		MPI_INTERCOMM_CREATE, NULL, 0, MPI_BYTE,
		remote_comm, &(remote_group->size), &d, &status);

    for (i = 0; i < remote_group->size; i++) {
	MPI_internal_rcv(peer_comm, remote_leader, MPI_INTERCOMM_CREATE,
		(char *)&MPI_nodes[MPI_num_nodes].listener_addr, 
		addr_size, MPI_BYTE, &d, &d, &d, &status);

        /* find if we already know about the address */
	for (j = 0; j < MPI_num_nodes; j++)
	   if (!memcmp(&MPI_nodes[j].listener_addr,
		       &MPI_nodes[MPI_num_nodes].listener_addr, addr_size))
	      break;
	if (j == MPI_num_nodes)
	    MPI_num_nodes++; /* new (unknown) node */
	remote_group->nodes[i] = j;
	remote_group->ranks[j] = i;
    }
#ifdef DEBUG
  printf("Group received from (rank,comm)=(%d,%d)\n",peer_comm, remote_leader);
#endif
}

/*========================================================================*/

void MPI_propagateGroup(MPI_Comm local_comm, int local_leader,
		     MPI_Group remote_group, MPI_Comm remote_comm, int high) {
  int i, j, d;
  int local_size = MPI_Comms[local_comm].group->size;
  size_t addr_size = sizeof(struct sockaddr_in);
  int node;

  for (i = 0; i < local_size; i++) {
    if (i == local_leader)  continue;
    MPI_internal_snd(local_comm, i, MPI_INTERCOMM_CREATE,
		NULL, 0, MPI_BYTE, MPI_first_available_comm, high, d);
    MPI_internal_snd(local_comm, i, MPI_INTERCOMM_CREATE,
		NULL, 0, MPI_BYTE,
		remote_comm, remote_group->size, d);
    for (j = 0; j < remote_group->size; j++) {
	node = remote_group->nodes[j];
	MPI_internal_snd(local_comm, i, MPI_INTERCOMM_CREATE,
	  (char *)&MPI_nodes[node].listener_addr, 
	  addr_size, MPI_BYTE, d, d, d);
    }
  }
}

/*========================================================================================*/

int MPI_Intercomm_merge(MPI_Comm intercomm, int high, MPI_Comm *newintracomm) {

  int i, d, tag = 0;
  MPI_Status status;
  struct MPI_Comm_entry entry, *new_entry;
  MPI_Group new_group;
  int real_high;
  int local_size, remote_size;
  
  if (MPI_first_available_comm >= MPI_MAX_COMMS) {
    printf("Can not allocate more communicators\n");
    exit(1);
  }
  if (intercomm == MPI_COMM_NULL) {
    printf("Error: trying to merge non communicator\n");
    exit(1);
  }
  
  entry = MPI_Comms[intercomm];
  if (entry.intracomm) {
    printf("Error: trying to merge intracomm\n");
    exit(1);
  }
  
  local_size = entry.local->size;
  remote_size = entry.group->size;
  
  if (entry.source == 0) { /* this is a local root */
    int remote_high, remote_node;
    MPI_Comm available;

#ifdef DEBUG
   printf("Root of %d entered merge\n", intercomm);
#endif

    /* find out max of available comm entries */
    for (i = 1; i < local_size; i++) {
      MPI_internal_rcv(-intercomm, i, MPI_INTERCOMM_MERGE, 
			NULL, 0, MPI_BYTE, &available, &d, &d, &status);
      if (available > MPI_first_available_comm)
    	MPI_first_available_comm = available;
    }

#ifdef DEBUG
   printf("Root of %d agread on new comm %d\n", intercomm, MPI_first_available_comm);
#endif
    
    /* sinchronize with remote root, 
       send/receive "high" and max available comm entry */
    if (entry.tag & MPI_HIGH_TAG) { /* to avoid deadlock */

printf("I AM HIGH!\n");

      MPI_internal_snd(intercomm, 0, MPI_INTERCOMM_MERGE, NULL, 0, MPI_BYTE,
       		MPI_first_available_comm, high, d);
      MPI_internal_rcv(intercomm, 0, MPI_INTERCOMM_MERGE, NULL, 0, MPI_BYTE,
         	&available, &remote_high, &d, &status);
    }
    else {
printf("I AM LOW!\n");
      MPI_internal_rcv(intercomm, 0, MPI_INTERCOMM_MERGE, NULL, 0, MPI_BYTE,
          	&available, &remote_high, &d, &status);
      MPI_internal_snd(intercomm, 0, MPI_INTERCOMM_MERGE, NULL, 0, MPI_BYTE,
           	MPI_first_available_comm, high, d);
    }
    
    /* calculate high flag and global entry */
    real_high = (high == remote_high) ? (MPI_node_num > remote_node) : high;
    if (available > MPI_first_available_comm)
      MPI_first_available_comm = available;
    
    /* send out global available comm entry and high flag */
    for (i = 1; i < local_size; i++) 
      MPI_internal_snd(-intercomm, i, MPI_INTERCOMM_MERGE, NULL, 0, MPI_BYTE, 
			MPI_first_available_comm, real_high, d);
  }
  else { /* non local root */
#ifdef DEBUG
   printf("Non-Root of %d entered merge\n", intercomm);
#endif

    /* send to local root my first_available comm entry */
    MPI_internal_snd(-intercomm, 0, MPI_INTERCOMM_MERGE, 
		NULL, 0, MPI_BYTE, MPI_first_available_comm, d, d);

    /* receive from local root global first_available comm entry and high flag */
    MPI_internal_rcv(-intercomm, 0, MPI_INTERCOMM_MERGE, 
		NULL, 0, MPI_BYTE, &MPI_first_available_comm, &real_high, &d, &status);
  }
  
  *newintracomm = MPI_first_available_comm++;
  new_entry = &MPI_Comms[*newintracomm];
  
  /* merge remote and local group */
  new_group = MPI_getGroup(); 
  if (real_high) { /* first local group then global */
    int node;
    struct MPI_Group_entry group_entry = *(entry.local);
    for (i = 0; i < local_size; i++) {
        node = group_entry.nodes[i];
        new_group->nodes[i] = node;
        new_group->ranks[node] = i;
    }
    group_entry = *(entry.group);
    for (i = 0; i < remote_size; i++) {
        node = group_entry.nodes[i];
        new_group->nodes[i+local_size] = node;
        new_group->ranks[node] = i+local_size;
    }
  }
  else { /* first remote group then local */
    int node;
    struct MPI_Group_entry group_entry = *(entry.group);
    for (i = 0; i < remote_size; i++) {
        node = group_entry.nodes[i];
        new_group->nodes[i] = node;
        new_group->ranks[node] = i;
    }
    group_entry = *(entry.local);
    for (i = 0; i < local_size; i++) {
        node = group_entry.nodes[i];
        new_group->nodes[i+remote_size] = node;
        new_group->ranks[node] = i+remote_size;
    }
  }
    
  new_group->size = local_size + remote_size;
  
  new_entry->group = new_group;
  new_group->references++;

  new_entry->group = new_group;
  new_group->references++;
  
  MPI_Group_free(&new_group);

  new_entry->snd_context = *newintracomm;
  new_entry->rcv_context = *newintracomm;
  new_entry->source = new_entry->group->ranks[MPI_node_num];
  new_entry->intracomm = 1;
  new_entry->used = 1;
  new_entry->tag = entry.tag;
  
  return MPI_SUCCESS;
}
/*========================================================================================*/
