#include "mpi.h"
#include "mpiimpl.h"

/*
struct MPI_Group_entry {
  MPI__INT nodes[MPI_MAX_NODES]; /* array of nodes' # ordered by the rank  
  MPI__INT ranks[MPI_MAX_NODES]; /* array of ranks ordered by the node #   
  MPI__INT size;
  MPI__INT references;
};
typedef struct MPI_Group_entry* MPI_Group;
*/

#define MAX_GROUPS 500

static struct MPI_Group_entry MPI_Groups[MAX_GROUPS];

MPI_Group MPI_GROUP_EMPTY;

void MPI_init_groups(void ) {
  int i;
  for (i = 0; i < MAX_GROUPS; i++)
    MPI_Groups[i].references = 0;
  
  MPI_GROUP_EMPTY = &MPI_Groups[0];
  MPI_GROUP_EMPTY->references = 1;
}

/*========================================================================================*/

int MPI_Group_size(MPI_Group group, int *size) {
  if (group->references < 1) {
    *size = MPI_UNDEFINED;
    return MPI_ERR_GROUP;
  }
  *size = group->size;
  return MPI_SUCCESS;
}

/*========================================================================================*/

int MPI_Group_rank(MPI_Group group, int *rank) {
  if (group->references < 1) {
    *rank = MPI_UNDEFINED;
    return MPI_ERR_GROUP;
  }
  *rank = group->ranks[MPI_node_num];
  return MPI_SUCCESS;
}


/*========================================================================================*/

int MPI_Group_free(MPI_Group *group) {
  if ((*group)->references < 1) {
    *group = MPI_GROUP_NULL;
    return MPI_ERR_GROUP;
  }
#ifdef _DEBUG
  printf("freed group with %d references\n", (*group)->references);
#endif  
  ((*group)->references)--;
#ifdef _DEBUG
  printf("freed group with %d more references\n", (*group)->references);
#endif  
  *group = MPI_GROUP_NULL;
  return MPI_SUCCESS;
}

/*========================================================================================*/

MPI_Group MPI_getGroup() {
  int entry = 1;
  int i;

  while ((MPI_Groups[entry].references > 0) &&
        (entry < MAX_GROUPS)) 
    entry++;

  if (entry == MAX_GROUPS) {/* no more available groups */
    printf("Can not allocate group\n");
    exit (1);
    return MPI_GROUP_NULL;
  }

  
  MPI_Groups[entry].size = 0;
#ifdef _DEBUG
  printf("getGroup returned entry = %d with %d ref\n", entry, MPI_Groups[entry].references);
#endif  
  MPI_Groups[entry].references = 1;
  for (i = 0; i < MPI_MAX_NODES; i++) {
    MPI_Groups[entry].nodes[i] = MPI_UNDEFINED;
    MPI_Groups[entry].ranks[i] = MPI_UNDEFINED;    
  }
#ifdef _DEBUG
  printf("getGroup returned entry = %d\n", entry);  
#endif  
  return &MPI_Groups[entry];
}

/*========================================================================================*/

int MPI_Group_union(MPI_Group group1, MPI_Group group2, MPI_Group *newgroup) {
  int i, rank, size, node;

  if (group1 == MPI_GROUP_NULL ||
      group2 == MPI_GROUP_NULL) {
    *newgroup = MPI_GROUP_NULL;
    return MPI_ERR_GROUP;
  }

  *newgroup = MPI_getGroup();

  size = group1->size;
  for (i = 0; i < size; i++) {
    node = group1->nodes[i];
    (*newgroup)->nodes[i] = node;
    (*newgroup)->ranks[node] = i;
  }

  rank = size;
  size = group1->size;
  for (i = 0; i < size; i++) {
    node = group2->nodes[i];
    if ((*newgroup)->ranks[node] != MPI_UNDEFINED) {
      (*newgroup)->nodes[rank] = node;
      (*newgroup)->ranks[node] = rank;
      rank++;
    }
  }
  if (rank == 0) {
    MPI_Group_free(newgroup);
    *newgroup = MPI_GROUP_EMPTY;
  }
  else
    (*newgroup)->size = rank;
  return MPI_SUCCESS;
}

/*========================================================================================*/

int MPI_Group_intersection(MPI_Group group1, MPI_Group group2, MPI_Group *newgroup) {
  int i, rank, size, node;

  if (group1 == MPI_GROUP_NULL ||
      group2 == MPI_GROUP_NULL) {
    *newgroup = MPI_GROUP_NULL;
    return MPI_ERR_GROUP;
  }

  *newgroup = MPI_getGroup();
  
  rank = 0;
  size = group1->size;
  for (i = 0; i < size; i++) {
    node = group1->nodes[i];
    if (group2->ranks[node] != MPI_UNDEFINED) {
      (*newgroup)->nodes[rank] = node;
      (*newgroup)->ranks[node] = rank;
      rank++;
    }      
  }
  
  if (rank == 0) {
    MPI_Group_free(newgroup);
    *newgroup = MPI_GROUP_EMPTY;
  }
  else
    (*newgroup)->size = rank;

  return MPI_SUCCESS;
}
/*========================================================================================*/

int MPI_Group_difference(MPI_Group group1, MPI_Group group2, MPI_Group *newgroup) {
  int i, rank, size, node;

  if (group1 == MPI_GROUP_NULL ||
      group2 == MPI_GROUP_NULL) {
    *newgroup = MPI_GROUP_NULL;
    return MPI_ERR_GROUP;
  }

  *newgroup = MPI_getGroup();
  
  rank = 0;
  size = group1->size;
  for (i = 0; i < size; i++) {
    node = group1->nodes[i];
    if (group2->ranks[node] == MPI_UNDEFINED) {
      (*newgroup)->nodes[rank] = node;
      (*newgroup)->ranks[node] = rank;
      rank++;
    }      
  }
  
  if (rank == 0) {
    MPI_Group_free(newgroup);
    *newgroup = MPI_GROUP_EMPTY;
  }
  else
    (*newgroup)->size = rank;

  return MPI_SUCCESS;
}
/*========================================================================================*/
int MPI_Group_incl(MPI_Group group, int n, int *ranks, MPI_Group *newgroup) {
  int i, node;

  if (group == MPI_GROUP_NULL) {
    *newgroup = MPI_GROUP_NULL;
    return MPI_ERR_GROUP;
  }

  *newgroup = MPI_getGroup();
  
  for (i = 0; i < n; i++) {
    node = group->nodes[ranks[i]];
    (*newgroup)->nodes[i] = node;
    (*newgroup)->ranks[node] = i;
  }

  if (n == 0) {
    MPI_Group_free(newgroup);
    *newgroup = MPI_GROUP_EMPTY;
  }
  else
    (*newgroup)->size = n;
  return MPI_SUCCESS;
}
/*========================================================================================*/

int MPI_Group_excl(MPI_Group group, int n, int *ranks, MPI_Group *newgroup) {
  int j, i, size, rank, node;

  if (group == MPI_GROUP_NULL) {
    *newgroup = MPI_GROUP_NULL;
    return MPI_ERR_GROUP;
  }
#ifdef DEBUG
  printf("entered group_excl\n");
#endif
  
  *newgroup = MPI_getGroup();
  rank = 0;
  size = group->size;

  for (i = 0; i < size; i++) {
    node = group->nodes[i];
    j = 0;
    while (j < n && ranks[j] != i) j++;
    if (j == n) {
      (*newgroup)->nodes[rank] = node;
      (*newgroup)->ranks[node] = rank;
      rank++;
    }
  }
  
  if (size - n == 0) {
    MPI_Group_free(newgroup);
    *newgroup = MPI_GROUP_EMPTY;
  }
  else
    (*newgroup)->size = size - n;

#ifdef DEBUG
  printf("exited group_excl\n");
#endif
  return MPI_SUCCESS;
}
/*========================================================================================*/
int MPI_Group_range_incl(MPI_Group group, int n, int ranges[][3], MPI_Group *newgroup);
/*========================================================================================*/
int MPI_Group_range_excl(MPI_Group group, int n, int ranges[][3], MPI_Group *newgroup);
/*========================================================================================*/
