#include "mpi.h"
#include "mpiimpl.h"

#include <stdio.h>
#include <sys/types.h>
#include <sys/wait.h>

#define HOSTNAME_LEN 256

static int is_initialized = 0;
static int is_finalized = 0;

/*========================================================================================*/
/*
  If this is call from slave, last argument has to be SLAVE_TAG.
  If this is call from spawned process last argument has to be SPAWNED_TAG.
  If none of the above is true this is call from master process.
*/
int MPI_Init(int *argc_ptr, char ***argv_ptr) { 
  int i; 

  if (is_initialized) {
    printf("MPI_Init:  can't call MPI_Init after it's already started.\n");
    exit(1);
  }

  if (sizeof(MPI__INT) != (size_t)4 ) {
    printf("sizeof(MPI__INT) != 4:  re-define MPI__INT in mpiimpl.h\n");
    exit(1);
  }

#ifdef DEBUG
  system("hostname");
  printf("entering MPI_Init\n");
  printf(" argc: %d\n", *argc_ptr);
  printf("last argv: %s\n", (*argv_ptr)[*argc_ptr - 1]);
#endif

  MPI_init_state();
 
  if (strcmp((*argv_ptr)[*argc_ptr - 1], SLAVE_TAG ) == 0) {
    if (MPI_slave(argc_ptr, argv_ptr)) /* this is a slave */
      return MPI_ERR_INTERN;
  }
  else if (strcmp((*argv_ptr)[*argc_ptr - 1], SPAWNED_TAG ) == 0) {
    if (MPI_spawned(argc_ptr, argv_ptr)) /* this is a spawned process */
      return MPI_ERR_INTERN;
  }
  else {  /* this is a master process */
    if (MPI_master(argc_ptr, argv_ptr))
      return MPI_ERR_INTERN;
  }

  if (MPI_init_rcv_buff(MPI_num_nodes))
    return MPI_ERR_INTERN;
  is_initialized = 1;

#ifdef DEBUG
  printf("Node %d initialized\n", MPI_node_num);
#endif

  return MPI_SUCCESS;
}

/*========================================================================================*/

int MPI_Initialized(int* flag) { 
  *flag = is_initialized;
  return MPI_SUCCESS;
}

/*========================================================================================*/

int MPI_Finalize() {
  int i, statusp;
  pid_t pid;

  if (is_finalized) {
    printf("MPI_Finalize: can't call MPI_Finaluze after it's already stoped.\n");
    exit(1);
  }

#ifdef DEBUG
    printf("waiting for %d slaves to finish, had %d nodes \n", MPI_slaves, MPI_nodes);
#endif

  /*close all open sockets for receiving */
  for (i = 0; i < MPI_num_nodes; i++)
    if (MPI_nodes[i].sd != MPI_NOSOCKET)
      shutdown(MPI_nodes[i].sd, 0);
  
  /*close listener for receiving */
    shutdown(MPI_my_list_sd, 0);
     
    for (i = 0; i < MPI_slaves; i++) {
      CALL_CHK(pid = wait, (&statusp));
#ifdef DEBUG
      printf("Child w/ pid =  %d status %d\n", pid, statusp);
#endif
    }
    
  is_finalized = 1;
  return MPI_SUCCESS;
}

/*========================================================================================*/

int MPI_Finalized(int *flag) {
  *flag = is_finalized;
  return MPI_SUCCESS;
}

/*========================================================================================*/
int MPI_Abort(MPI_Comm comm, int errorcode) { 
    printf("MPI_Abort called with errorcode: %d\n", errorcode);
    abort();
    return MPI_SUCCESS; /* return for lint */
}
