#include "mpiimpl.h"

#include <sys/stat.h> 
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/socket.h>

/* Extract and erase procgroup name from argument list */

#define PROCGROUP_FILENAME "procgroup"
#define SET_PROCGROUP_FILENAME_TAG "-p4pg"

char* MPI_procgroup_file(int *argc_ptr, char ***argv_ptr) {

  char* pg_file = PROCGROUP_FILENAME;
  int set_pg_new_name = 0;
  int i;

  for(i = 0; i < *argc_ptr; i++ ) {
    if (set_pg_new_name)
      /* set tag name_is_found, move rest of the args for 2 places */
      (*argv_ptr)[i] = (*argv_ptr)[i+2];
    else 
      if (!strcmp((*argv_ptr)[i], SET_PROCGROUP_FILENAME_TAG)) {
	pg_file = (*argv_ptr)[i+1]; /* get procgroup file name */
	*argc_ptr -= 2;             /* ignore two arguments    */
	set_pg_new_name = 1;
	i--;
      }
  }
  if (set_pg_new_name)
    (*argv_ptr)[*argc_ptr] = NULL;
  
  { /* Check status of the procgroup file */
    struct stat buf;
    if ( 0 != stat(pg_file, &buf) || 
	 !S_ISREG(buf.st_mode) /*!(buf.st_mode & S_IFREG)--not POSIX*/ ||
	 !(buf.st_mode & S_IRUSR) ) {
/*
      fprintf(stderr, "MPI:  can't read procgroup file:  %s\n", pg_file );
      if (0 != strcmp(pg_file, PROCGROUP_FILENAME))
	fprintf(stderr,"  Default path is current directory.\n");
      fprintf(stderr,"  Alternatively, try adding a command line arg:"
	      "  -p4pg  PATH_TO_PROCGROUP_FILE\n");
*/
      return NULL;  /* do not FAIL */
    }
  }
  return pg_file;
}

/*========================================================================================*/
/* Get a new listener, and port for this process. All arguments are output. */

int MPI_new_listener(int *sd,                 /* bound socket descriptor */
		     struct sockaddr_in *sin, /* inet info structure */
		     int *port                /* port number of bound sd*/) {

  char host[256];       /* hostname for this process      */
  struct hostent *hp;   /* ptr to host info structure     */
  int i;

  /* holders for various type sizes that can be pull out of the loops */
  size_t addr_size = sizeof(struct sockaddr_in);

  /* use the default protocol 6 = IPPROTO_TCP */
  *sd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
  MPI_new_sd(*sd);

  hp = gethostbyname(MPI_processor_name);
  bzero((char*)sin, addr_size);

  /* Is this next line (h_addr) needed? */
  memcpy((char *)&(sin->sin_addr), hp->h_addr, hp->h_length);
  
  sin->sin_port = htons(0); /* get any available port */
  sin->sin_family = hp->h_addrtype;
  /* Since INADDR_ANY is 0, zeroing out sin accomplished this, anyway. */
  sin->sin_addr.s_addr = INADDR_ANY;

  CALL_CHK(bind, (*sd, (struct sockaddr *)sin, addr_size));

  /* i acts as dummy for sin_len below */
  i = addr_size;
  CALL_CHK(getsockname, (*sd, (struct sockaddr *)sin, &i));
  *port = ntohs(sin->sin_port);

  if (*port == 0) {
    printf("Listener failed to get new port!!\n");
    exit(1);
  }

  /* Under SunOS 4.1, SOMAXCONN = 5; Could have slaves re-try after timeout */
  CALL_CHK(listen, (*sd, SOMAXCONN));
  /* On Solaris 2.6, getsockname (and maybe bind?) zero out sin->sin_addr
    (converting it to localhost for efficiency?); So copy it back in */
  memcpy((char *)&(sin->sin_addr), hp->h_addr, hp->h_length);
  return 0;
}

/*========================================================================================*/
/* read procgroup array and create slave processes */

int MPI_exec_pg_file(char *hostname, char *port, 
		     int argc, char *argv[]) {
  int i;  
  for(i = 1; i < MPI_MAX_NODES && MPI_pg_array[i].processor != NULL; i++)
    MPI_fork_process(MPI_pg_array[i].processor, MPI_pg_array[i].process,
      		     argc, argv, /* arguments excluding command */
		     hostname, port, SLAVE_TAG);
    
  return i-1; /* return number of slaves */
}

/*========================================================================================*/

#define PROCGROUP_LEN 10000
#define BUF_LEN 256

/* parse_buf must be as large as procgroup file length; avoids using malloc */
char parse_buf[PROCGROUP_LEN];

/* reads pg_file and populates MPI_pg_array */

int MPI_parse(char* pg_file ) {
  FILE *fin;
  char *buf = parse_buf;
  char *t0;
  int index;
  
  MPI_pg_array[0].processor = "local";
  MPI_pg_array[0].num_threads = "0";
  MPI_pg_array[0].process = ""; /* Could use strcat of all argv[i] ?? */
  index = 1;

  if (fin = fopen(pg_file, "r")) { 
    while(fgets(buf, BUF_LEN, fin)) {
      if (buf[0] == '#' ) 
	continue; /* # means comment */
      if (buf[0] == '\n') 
	continue; /* skip null lines */
      
      t0 = strtok(buf, " ");
      if (!strcmp(t0, "local"))
        continue;
      else {
	MPI_pg_array[index].processor = t0;
	if ((t0 = strtok(NULL, " ")) == NULL) {
	  printf("parse: strtok: token not found\n");
	  fclose(fin);
	  exit(1);
	}
	MPI_pg_array[index].num_threads = t0;
	if ((t0 = strtok(NULL, "\n")) == NULL) {
	  printf("parse: strtok: token not found\n");
	  fclose(fin);
	  exit(1);
	}
	MPI_pg_array[index].process = t0;
	buf = strchr(t0, '\0') + 1;
      }
      index++;
    }
    fclose(fin);
  }
  return 0;
}

/*========================================================================================*/

void MPI_print_warning1() {
  printf("WARNING:  This architecture allows only %d connections before\n", SOMAXCONN);
  printf(" connections are dropped from the queue.  There are %d slaves.\n", MPI_slaves);
  printf(" Some connections may be lost.\n");
}

