#include "mpi.h"
#include "mpiimpl.h"

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/socket.h>
#include <sys/stat.h> 

/*======================================================================*/

int MPI_master(int *argc, char ***argv) { /* original input arguments */
  int i, j, ns;                    /* socket descriptor            */
  struct MPI_init_msg msg1;
  char new_port[8];             /* port on which master listens, ASCII number */
  size_t fromlen; 
  int node_num;

  struct hostent *hp;           /* ptr to host info structure     */
  struct sockaddr_in *sin;      /* inet info structure            */
  struct sockaddr_in new_sin;   /* used for each new connection   */

  /* holders for type sizes that can be pull out of the loops */
  size_t msg1_size = sizeof(struct MPI_init_msg);
  size_t addr_size = sizeof(struct sockaddr_in);
  
  char *pg_file = MPI_procgroup_file(argc, argv);

  MPI_node_num = 0; /* master is process with number 0 */

  hp = gethostbyname(MPI_processor_name);

#ifdef DEBUG
  printf("Master's hostname is: %s\n", MPI_processor_name);
#endif

  sin = &(MPI_nodes[MPI_node_num].listener_addr);
  /* open new socket and bind it to port */
  MPI_new_listener(&MPI_my_list_sd, sin, &MPI_my_port); 

  MPI_my_port = ntohs(sin->sin_port);
  sprintf(new_port, "%d", MPI_my_port);

#ifdef DEBUG
  printf("Master now listens at port %d with socket %d\n", MPI_my_port, MPI_my_list_sd);
#endif
  
  if (pg_file != NULL) { /* create singlton MPI process */
	MPI_parse(pg_file);      /* read the procgroup file */
 	 /* crate slaves by executing info from procgroup file */
  	MPI_slaves += MPI_exec_pg_file(hp->h_name, new_port, *argc-1, *argv + 1);
  }

  if (MPI_slaves > SOMAXCONN) 
    MPI_print_warning1(); /* more slave then max places in connection queue */

  MPI_num_nodes = MPI_slaves + 1;
  if (MPI_num_nodes < 1)
    return MPI_ERR_INTERN; /* FAIL How this happened ? */

#ifdef DEBUG
  printf("Number of nodes is %d \n", MPI_nodes);
#endif
  
  /* establish connection with the slaves */
  for (node_num = 1; node_num <= MPI_slaves; node_num++) { 
    fromlen = addr_size;
    CALL_CHK(ns = accept, (MPI_my_list_sd, (struct sockaddr *)&new_sin, &fromlen));

#ifdef DEBUG
    printf("Master accepted socket %d for node %d\n", ns, node_num);
#endif

    MPI_nodes[node_num].sd = ns;
    MPI_new_sd(ns);
    
    /* receive slave's listener_addr and store in field of pg_array[rank] */
    CALL_CHK(recv, (ns, (char *)&(MPI_nodes[node_num].listener_addr),
			addr_size, 0));

    /* send slaves rank and total number of nodes */
    msg1.len = htonl(msg1_size);
    msg1.node_num = htonl(node_num);
    msg1.num_nodes = htonl(MPI_num_nodes);
    CALL_CHK(send, (ns, (char *)&msg1, msg1_size, 0));
  }
  
   /* Send listener_addr fields of every node to slave */
    for (i = 1; i <= MPI_slaves; i++) 
      for (j = 0; j < MPI_num_nodes; j++ )
	CALL_CHK(send, (MPI_nodes[i].sd,
			(char*)&MPI_nodes[j].listener_addr, 
			addr_size, 0));

  { /* Acknowledge that master and slave are synchronized */
    MPI__INT buf;
    buf =  htonl(addr_size);
    for (i = 1; i <= MPI_slaves; i++ ) {
      CALL_CHK(send, (MPI_nodes[i].sd, (char *)&buf, sizeof(buf), 0));
#ifdef DEBUG
      printf("Master and node %d are synchronized\n", i);
#endif
      /*MPI_pg_array[i].sd = MPI_PG_NOSOCKET; */
    }
  }

  MPI_createSelfComm();
  MPI_createWorldComm(MPI_num_nodes);
  return MPI_SUCCESS;
}

void MPI_createSelfComm() {
  int i;
  MPI_Group group;
  struct MPI_Comm_entry *entry;

  if (MPI_COMM_SELF != MPI_COMM_NULL)
	return;

  MPI_COMM_SELF = MPI_first_available_comm++;
  
  group = MPI_getGroup();
  group->ranks[0] = MPI_node_num;
  group->nodes[MPI_node_num] = 0;
  group->size = 1;
  
  entry = &MPI_Comms[MPI_COMM_SELF];
  entry->group = group;
  group->references++;

  entry->local = group;
  group->references++;

  MPI_Group_free(&group);

  entry->snd_context = MPI_COMM_SELF;
  entry->rcv_context = MPI_COMM_SELF;
  
  entry->source = 0;
  entry->intracomm = 1;
  entry->used = 1;
  entry->tag = 0;
}

void MPI_createWorldComm(int num_nodes) {
  int i;
  MPI_Group group;
  struct MPI_Comm_entry *entry;

  if (MPI_COMM_WORLD != MPI_COMM_NULL)
	return;
  MPI_COMM_WORLD = MPI_first_available_comm++;
  
  group = MPI_getGroup();
  for (i = 0; i < num_nodes; i++) {
    group->ranks[i] = i;
    group->nodes[i] = i;
  }
  group->size = num_nodes;
  
  entry = &MPI_Comms[MPI_COMM_WORLD];
  entry->group = group;
  group->references++;

  entry->local = group;
  group->references++;

  MPI_Group_free(&group);

  entry->snd_context = MPI_COMM_WORLD;
  entry->rcv_context = MPI_COMM_WORLD;
  
  entry->source = MPI_node_num;
  entry->intracomm = 1;
  entry->used = 1;
  entry->tag = 0;
}
