/* This is public portion of mpi.h */
#ifndef MPI_H

#define MPI_H

/* Keep C++ compilers from getting confused */
#if defined(__cplusplus)
extern "C" {
#endif

#include <stddef.h>

/* INT must be 32-bit quantity.  Re-define to long if necessary. */
#if (1 << 31) == 0
#define MPI__INT long
#else
#define MPI__INT int
#endif
  
#define MPI_MAX_NODES  1000

/* Elementary datatypes:  */
typedef MPI__INT MPI_Datatype;

#define MPI_CHAR           ((MPI_Datatype)1)
#define MPI_SHORT          ((MPI_Datatype)2)
#define MPI_INT            ((MPI_Datatype)3)
#define MPI_LONG           ((MPI_Datatype)4)
#define MPI_UNSIGNED_CHAR  ((MPI_Datatype)5)
#define MPI_UNSIGNED_SHORT ((MPI_Datatype)6)
#define MPI_UNSIGNED       ((MPI_Datatype)7)
#define MPI_UNSIGNED_LONG  ((MPI_Datatype)8)
#define MPI_FLOAT          ((MPI_Datatype)9)
#define MPI_DOUBLE         ((MPI_Datatype)10)
#define MPI_LONG_DOUBLE    ((MPI_Datatype)11)
#define MPI_BYTE           ((MPI_Datatype)12)
#define MPI_LONG_LONG_INT  ((MPI_Datatype)13)

/* Implementation of this should go together*/
#if 0
#define MPI_PACKED         ((MPI_Datatype)14)
int MPI_Pack(void *inbuf, int incount, MPI_Datatype datatype, 
         void *outbuf, int outsize, int *position, MPI_Comm comm);
int MPI_Unpack(void *inbuf, int insize, int *position, 
           void *outbuf, int outcount, MPI_Datatype datatype, MPI_Comm comm);
int MPI_Packsize(int incount, MPI_Datatype, MPI_Comm comm, int *size);
#endif

/* handles to assorted structures */

typedef struct { 
  int MPI_SOURCE;
  int MPI_TAG;
  int MPI_ERROR;
  size_t mpi_size;
} MPI_Status;

typedef MPI__INT MPI_Comm;

/* predifined communicators */
extern MPI_Comm MPI_COMM_WORLD;
extern MPI_Comm MPI_COMM_SELF;

struct MPI_Group_entry {
  MPI__INT nodes[MPI_MAX_NODES]; /* array of nodes  ordered by the rank         */
  MPI__INT ranks[MPI_MAX_NODES]; /* array of ranks ordered by the node number   */

  MPI__INT size;
  MPI__INT references;
};

typedef struct MPI_Group_entry* MPI_Group;


#if 0
typedef /* something here */ MPI_Request;
#endif

/* Null handles */ 
#define MPI_COMM_NULL       ((MPI_Comm) -2)
#define MPI_DATATYPE_NULL   ((MPI_Datatype) 0)
#define MPI_GROUP_NULL      ((MPI_Group) NULL)
extern MPI_Group MPI_GROUP_EMPTY;

#if 0
#define MPI_REQUEST_NULL 
#define MPI_OP_NULL 
#define MPI_ERRHANDLER_NULL 
#endif

/* assorted constants*/ 
/* 
#define MPI_BOTTOM  
*/
#define MPI_PROC_NULL -1 
#define MPI_ANY_SOURCE -1
#define MPI_ANY_TAG -1
#define MPI_UNDEFINED -1
/*
#define MPI_BSEND_OVERHEAD
#define MPI_KEYVAL_INVALID 
*/

/* environmental inquiry keys */ 
#define MPI_HOST 1
#define MPI_TAG_UB 2
#define MPI_IO 3
#define MPI_WTIME_IS_GLOBAL 4
#define MPI_UNIVERSE_SIZE

/* From MPI-1.2 */
#define MPI_VERSION 1
#define MPI_SUBVERSION 2

/* return codes */ 
#define MPI_SUCCESS 0
#define MPI_ERR_BUFFER 1
#define MPI_ERR_COUNT 2
#define MPI_ERR_TYPE 3
#define MPI_ERR_TAG 4
#define MPI_ERR_COMM 5
#define MPI_ERR_RANK  6
#define MPI_ERR_REQUEST 7
#define MPI_ERR_ROOT 8
#define MPI_ERR_GROUP 9
#define MPI_ERR_OP 10
#define MPI_ERR_TOPOLOGY 11
#define MPI_ERR_DIMS 12
#define MPI_ERR_ARG 13
#define MPI_ERR_UNKNOWN 14
#define MPI_ERR_TRUNCATE 15
#define MPI_ERR_OTHER 16
#define MPI_ERR_INTERN 17
#define MPI_PENDING 18
#define MPI_ERR_IN_STATUS 19
#define MPI_ERR_LASTCODE 20

#define MPI_ERR_SPAWN 21
#define MPI_ERR_PORT 21


/* Error-handling specifiers */ 
#define MPI_ERRORS_ARE_FATAL 1
#define MPI_ERRORS_RETURN 2


/* Maximum sizes for strings */ 
#define MPI_MAX_PROCESSOR_NAME  256
#define MPI_MAX_ERROR_STRING    256

/* collective operations */ 
#if 0
#define MPI_MAX 
#define MPI_MIN 
#define MPI_SUM 
#define MPI_PROD 
#define MPI_MAXLOC 
#define MPI_MINLOC 
#define MPI_BAND 
#define MPI_BOR 
#define MPI_BXOR 
#define MPI_LAND 
#define MPI_LOR 
#define MPI_LXOR 
#endif

/* prototypes for user-defined functions (C) */ 
typedef int  MPI_Copy_function(MPI_Comm oldcomm, int keyval, void *extra_state, 
                   void *attribute_val_in, void *attribute_val_out, int *flag); 
typedef int  MPI_Delete_function(MPI_Comm comm, int keyval, void *attribute_val, 
                 void *extra_state);
typedef void MPI_Handler_function(MPI_Comm *, int *, ...);
typedef void MPI_User_function( void *invec, void *inoutvec, 
                int *len, MPI_Datatype *datatype); 

/* Function declarations*/

/* Environmental Inquiry */
int MPI_Init(int *argc_ptr, char ***argv_ptr);
int MPI_Finalize(void);
int MPI_Abort(MPI_Comm comm, int errorcode);

/* Environmental Inquiry */
int MPI_Initialized(int *flag);
int MPI_Finalized(int *flag);   /* from MPI-2 */
int MPI_Get_processor_name(char *name, int *resultlen);
int MPI_Get_version(int *version, int *subversion);
int MPI_Error_string(int errorcode, char *string, int *resultlen);
int MPI_Comm_get_parent(MPI_Comm *parent);

int MPI_Get_count(MPI_Status *status, MPI_Datatype datatype, int *count);

/* Point to point communication */
int MPI_Iprobe(int source, int tag, MPI_Comm comm, int *flag, MPI_Status *status);
int MPI_Probe(int source, int tag, MPI_Comm comm, MPI_Status *status);

int MPI_Recv(void *buf, int count, MPI_Datatype datatype, int source, int tag,
         MPI_Comm comm, MPI_Status *status);
int MPI_Send(void *buf, int count, MPI_Datatype datatype, int dest, int tag, MPI_Comm comm);

/*Buffered send*/
#if 0
int MPI_Bsend(boid *buf, int count, MPI_Datatype datatype, int dest, int tag, MPI_Comm comm);
int MPI_Buffer_atach(void *buffer, int size);
int MPI_Buffer_detach(void *buffer, int size);
#endif

/* Other sends */
#if 0
int MPI_Ssend(boid *buf, int count, MPI_Datatype datatype, int dest, int tag, MPI_Comm comm);
int MPI_Rsend(boid *buf, int count, MPI_Datatype datatype, int dest, int tag, MPI_Comm comm);
#endif
  
/* Collective communication */
int MPI_Barrier(MPI_Comm comm);
int MPI_Bcast(void* buffer, int count, MPI_Datatype datatype, int root, MPI_Comm comm );
int MPI_Gather(void* sendbuf, int sendcount, MPI_Datatype sendtype, 
	       void* recvbuf, int recvcount, MPI_Datatype recvtype, 
	       int root, MPI_Comm comm);
int MPI_Gatherv(void* sendbuf, int sendcount, MPI_Datatype sendtype, 
		void* recvbuf, int *recvcounts, int *displs, MPI_Datatype recvtype, 
		int root, MPI_Comm comm);
int MPI_Scatter(void* sendbuf, int sendcount, MPI_Datatype sendtype, 
		void* recvbuf, int recvcount, MPI_Datatype recvtype, 
		int root, MPI_Comm comm);
int MPI_Scatterv(void* sendbuf, int *sendcounts, int *displs, MPI_Datatype sendtype, 
		 void* recvbuf, int recvcount, MPI_Datatype recvtype, 
		 int root, MPI_Comm comm);

#if 0
int MPI_Allgather(void* sendbuf, int sendcount, MPI_Datatype sendtype, 
		  void* recvbuf, int recvcount, MPI_Datatype recvtype, 
		  MPI_Comm comm);
int MPI_Allgatherv(void* sendbuf, int sendcount, MPI_Datatype sendtype, 
		   void* recvbuf, int *recvcounts, int *displs, MPI_Datatype recvtype, 
		   MPI_Comm comm);
int MPI_Alltoall(void* sendbuf, int sendcount, MPI_Datatype sendtype, 
		 void* recvbuf, int recvcount, MPI_Datatype recvtype, 
		 MPI_Comm comm);
int MPI_Alltoallv(void* sendbuf, int *sendcounts, int *sdispls, MPI_Datatype sendtype, 
		  void* recvbuf, int *recvcounts, int *rdispls, MPI_Datatype recvtype, 
		  MPI_Comm comm);
int MPI_Reduce(void* sendbuf, void* recvbuf, int count, MPI_Datatype datatype, 
	       MPI_Op op, int root, MPI_Comm comm);
int MPI_Op_create(MPI_User_function *function, int commute, MPI_Op *op);
int MPI_Op_free(MPI_Op *op);
int MPI_Allreduce(void* sendbuf, void* recvbuf, int count, MPI_Datatype datatype, 
		  MPI_Op op, MPI_Comm comm);
int MPI_Reduce_scatter(void* sendbuf, void* recvbuf, int *recvcounts, 
		       MPI_Datatype datatype, MPI_Op op, MPI_Comm comm);
int MPI_Scan(void* sendbuf, void* recvbuf, int count, MPI_Datatype datatype, 
	     MPI_Op op, MPI_Comm comm);
#endif

/* Group operations */
int MPI_Group_size(MPI_Group group, int *size);
int MPI_Group_rank(MPI_Group group, int *rank);
int MPI_Group_translate_ranks (MPI_Group group1, int n, int *ranks1, 
                   MPI_Group group2, int *ranks2);
int MPI_Group_compare(MPI_Group group1,MPI_Group group2, int *result);
int MPI_Comm_group(MPI_Comm comm, MPI_Group *group);

int MPI_Group_union(MPI_Group group1, MPI_Group group2, MPI_Group *newgroup);
int MPI_Group_intersection(MPI_Group group1, MPI_Group group2, MPI_Group *newgroup);
int MPI_Group_difference(MPI_Group group1, MPI_Group group2, MPI_Group *newgroup);
int MPI_Group_incl(MPI_Group group, int n, int *ranks, MPI_Group *newgroup);
int MPI_Group_excl(MPI_Group group, int n, int *ranks, MPI_Group *newgroup);
#if 0
int MPI_Group_range_incl(MPI_Group group, int n, int ranges[][3], MPI_Group *newgroup);
int MPI_Group_range_excl(MPI_Group group, int n, int ranges[][3], MPI_Group *newgroup);
#endif 
int MPI_Group_free(MPI_Group *group);

/* Communicator operations */
int MPI_Comm_size(MPI_Comm comm, int *size);
int MPI_Comm_rank(MPI_Comm comm, int *rank);
int MPI_Comm_compare(MPI_Comm comm1,MPI_Comm comm2, int *result);
int MPI_Comm_dup(MPI_Comm comm, MPI_Comm *newcomm);
int MPI_Comm_create(MPI_Comm comm, MPI_Group group, MPI_Comm *newcomm);
int MPI_Comm_split(MPI_Comm comm, int color, int key, MPI_Comm *newcomm);
int MPI_Comm_free(MPI_Comm *comm);
int MPI_Comm_test_inter(MPI_Comm comm, int *flag);
int MPI_Comm_remote_size(MPI_Comm comm, int *size);
int MPI_Comm_remote_group(MPI_Comm comm, MPI_Group *group);
int MPI_Intercomm_create(MPI_Comm local_comm, int local_leader, 
             MPI_Comm peer_comm, int remote_leader, int tag,
             MPI_Comm *newintercomm);
int MPI_Intercomm_merge(MPI_Comm intercomm, int high, MPI_Comm *newintracomm);

#if 0
int MPI_Keyval_create(MPI_Copy_function *copy_fn, MPI_Delete_function *delete_fn, 
              int *keyval, void* extra_state);
int MPI_Keyval_free(int *keyval);
int MPI_Attr_put(MPI_Comm comm, int keyval, void* attribute_val);
int MPI_Attr_delete(MPI_Comm comm, int keyval);
#endif
int MPI_Attr_get(MPI_Comm comm, int keyval, void* attribute_val, int *flag);

/* Nonblocking communication*/
#if 0  
int MPI_Request_free(MPI_Request *request);
int MPI_ISend(void *buf, int count, MPI_Datatype datatype, 
	      int dest, int tag, MPI_Comm comm, MPI_Request *request);
int MPI_IBSend(void *buf, int count, MPI_Datatype datatype, 
	       int dest, int tag, MPI_Comm comm, MPI_Request *request);
int MPI_ISSend(void *buf, int count, MPI_Datatype datatype, 
	       int dest, int tag, MPI_Comm comm, MPI_Request *request);
int MPI_IRSend(void *buf, int count, MPI_Datatype datatype, 
	       int dest, int tag, MPI_Comm comm, MPI_Request *request);
int MPI_IRecv(void *buf, int count, MPI_Datatype datatype, 
	      int source, int tag, MPI_Comm comm, MPI_Request *request);
int MPI_Cancel(MPI_Request *request);
int MPI_Canceled(MPI_Status *status, int* flag);
int MPI_Wait(MPI_Request *request, MPI_Status *status);
int MPI_Test(MPI_Request *request, int *flag, MPI_Status *status);
int MPI_Waitany(int count, MPI_Request *array_of_requests, 
		int *index, MPI_Status *status);
int MPI_Testany(int count, MPI_Request *array_of_requests, 
		int *index, int *flag, MPI_Status *status);
int MPI_Waitall(int count, MPI_Request *array_of_requests, 
		MPI_Status *array_of_statuses);
int MPI_Testany(int count, MPI_Request *array_of_requests, 
		int *flag, MPI_Status *array_of_statuses);
int MPI_Waitsome(int incount, MPI_Request *array_of_requests, 
		 int *outcount, int* array_of_indices, MPI_Status *array_of_statuses);
int MPI_Testsome(int incount, MPI_Request *array_of_requests, 
		 int *outcount, int* array_of_indices, MPI_Status *array_of_statuses);
#endif

typedef int MPI_Info;/* this is ignored for now, no info functions implemented */
#define MPI_INFO_NULL ((MPI_Info) 0)
#define MPI_ARGV_NULL ((char **) NULL)
#define MPI_ARGVS_NULL ((char ***) NULL)
#define MPI_ERRCODES_IGNORE ((int *) NULL)
 
int MPI_Comm_spawn(char *command, char *argv[], int maxprocs,
		   MPI_Info info, int root, MPI_Comm comm, 
		   MPI_Comm *intercomm, int array_of_errcodes[]);

#define MPI_MAX_PORT_NAME (MPI_MAX_PROCESSOR_NAME + 8)
int MPI_Open_port(MPI_Info info, char *port_name);
int MPI_Close_port(char *port_name);
int MPI_Comm_accept(char *port_name, MPI_Info info, int root,
		MPI_Comm comm, MPI_Comm *newcomm);
int MPI_Comm_connect(char *port_name, MPI_Info info, int root,
		MPI_Comm comm, MPI_Comm *newcomm);

#if defined(__cplusplus)
}
#endif

#endif /* #ifdef MPI_H */
