#include "mpi.h"

#define unDEBUG

#include <errno.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/select.h>

/* CALL-CHK tryes to call the function until it succeds.
   If function call is interupted by signal (errno == EINTR)
   it tryes again, and if the call failed for any other
   reson exits the program.
   In GNU library there is TEMP_FAILURE_RETRY
   macro that can be used instead of CALL_CHK */
#define CALL_CHK(function,args)   while (1) {     /* On EINTR (interrupted by signal), just try again */ if (((function args) == -1) &&     (errno != EINTR)) {      printf("Error: CALL_CHK err. number %d at %d\n", errno, MPI_node_num);     exit(1);  }   else break; /* successful system call */   }

/* special message tags */
#define MPI_P2P                   ((MPI__INT) -100)
#define MPI_INIT                  ((MPI__INT) -101) /* -1 is ANY_TAG */
#define MPI_BARRIER               ((MPI__INT) -102)
#define MPI_COMM_DUP              ((MPI__INT) -103)
#define MPI_COMM_CREATE           ((MPI__INT) -104)
#define MPI_COMM_SPLIT            ((MPI__INT) -105)
#define MPI_INTERCOMM_CREATE      ((MPI__INT) -106)
#define MPI_INTERCOMM_MERGE       ((MPI__INT) -107)
#define MPI_SEND_GROUP            ((MPI__INT) -108)
#define MPI_BCAST                 ((MPI__INT) -109)
#define MPI_GATHER                ((MPI__INT) -110)
#define MPI_GATHERV               ((MPI__INT) -111)
#define MPI_SCATTER               ((MPI__INT) -12)
#define MPI_SCATTERV              ((MPI__INT) -113)
#define MPI_COMM_SPAWN            ((MPI__INT) -114)
#define MPI_ACCEPT                ((MPI__INT) -115)


extern int MPI_master(int *argc_ptr, char ***argv_ptr);
extern int MPI_slave(int *argc_ptr, char ***argv_ptr);
extern int MPI_spawned(int *argc_ptr, char ***argv_ptr);

extern void MPI_initPorts(void);
extern void MPI_init_state(void);
extern int MPI_init_rcv_buff(int initial_nodes);

struct MPI_msg_hdr {
  MPI_Comm comm;
  MPI__INT source;   /* process rank */
  MPI__INT tag; 	/* if less then zero this is special msg 
			   otherwise it is some point to point message */
  MPI_Datatype datatype;
  MPI__INT size;     /* size is in bytes */

  MPI__INT info1;    /* additional fields for special use */
  MPI__INT info2;
  MPI__INT info3;
};
extern void *MPI_put_rcv_buff(struct MPI_msg_hdr *msg_hdr);
extern int MPI_peek_rcv_buff(MPI_Comm comm, int tag, int source, struct MPI_msg_hdr *hdr);
extern int MPI_get_rcv_buff(MPI_Comm comm, int tag, int source, int size, void *buf, struct MPI_msg_hdr *hdr);

#define MPI_NOT_BLOCKING 0
#define MPI_BLOCKING 1

#define MPI_DONT_EAT_HDR 0
#define MPI_EAT_HDR 1

#define MPI_NO_MSG_WAITING 0
#define MPI_MSG_WAITING 1

extern int MPI_read_msg_hdr(int source, int tag, MPI_Comm comm, 
                int *flag, struct MPI_msg_hdr *msg_hdr, int *fd,
                int blocking_flag, int eat_hdr__flag);

extern int MPI_internal_rcv(MPI_Comm comm, int source, int tag, void* buf, int size,
                MPI_Datatype datatype, 
                MPI__INT *info1, MPI__INT *info2,MPI__INT *info3,
                MPI_Status *status);
extern int MPI_internal_snd(MPI_Comm comm, int dest, int tag, void* buf, int size,
                MPI_Datatype datatype, 
                MPI__INT info1, MPI__INT info2, MPI__INT info3);

extern int MPI_is_connected(int socket); /* this shouldn't be used, it is slow
                        better try to send and catch SIGPIPE if broken */
extern void MPI_new_sd(int sd);

extern void MPI_init_groups(void);
extern MPI_Group MPI_getGroup(void);

#define MPI_PACK_GROUP_SIZE (sizeof(MPI__INT)*(2*MPI_MAX_NODES +1))
extern void MPI_pack_group(MPI_Group group, char *buff);
extern void MPI_unpack_group(char *buff, MPI_Group *group);

extern void MPI_propagateGroup(MPI_Comm local, int root, MPI_Group group,
			MPI_Comm groups_comm, int high_tag);

extern char* MPI_procgroup_file(int *argc_ptr, char ***argv_ptr);
extern int MPI_parse(char* pg_file);

extern int MPI_exec_pg_file(char *hostname, char *port, int argc, char *argv[]);
extern void MPI_print_warning1();

#define SLAVE_TAG "-p4amslave"
#define SPAWNED_TAG "-spawn"

extern pid_t MPI_fork_process(char *processor, char *process, 
                  int argc, char *argv[],
                  char *hostname, char *port, char *tag);

extern void MPI_createSelfComm(void);
extern void MPI_createWorldComm(int size);
/*======================================================================*/


#define MPI_NOSOCKET -1 /* must not coincide with socket descriptor int */

extern int MPI_node_num;     /* identifier of this process */
extern int MPI_slaves;       /* number of slaves this process has created/spawned,
		                at the end it has to wait for them, to finish     */
extern int MPI_num_nodes;        /* current number of known nodes */

extern int MPI_my_port;      /* important only at master node                     */
extern int MPI_my_list_sd;   /* socket where node listens for connections         */

extern fd_set MPI_fdset;
extern int    MPI_max_sd;    /* max value of socket descriptor -- useful with select() */

extern char MPI_processor_name[];
extern int MPI_processor_name_len;

extern size_t MPI_datatype_size[];

#define MPI_PID_UNAVAILABLE ((pid_t) -2)
#define MPI_PID_DIED        ((pid_t) -1)

struct MPI_node_struct {
  int sd; /* socket descriptor */
  struct sockaddr_in listener_addr; /* listener port to create new sockets */
  pid_t pid;
};
	
struct MPI_pg_struct {
  char *processor;
  char *num_threads;
  char *process;
};

/* index into pg_array is same as process number */
extern struct MPI_pg_struct MPI_pg_array[];
extern struct MPI_node_struct MPI_nodes[];

/* for more info look at MPI_1.1 report Ch.5 pg.156 */
struct MPI_Comm_entry {
  MPI_Group group;      /* (remote) group of the communicator */
  MPI_Group local;      /* local group in case of inter-communicators 
                           local==grop for intra-communicators                      */
  MPI_Comm snd_context; /* MPI_Comm to send, useful for inter-communicators         */
  MPI_Comm rcv_context; /* it is same as an index of this struct in MPI_Comms array 
                           rcv_context==snd_context for intra-communicators,
               rcv_context is local comm and snd_context is remote comm */
  
  int source;           /* rank of the process in (local) group                     */
  int intracomm;        /* "Is intracommunicator ?" flag                            */ 
  int used;
  int tag;		/* currently unused 	*/
};
#define MPI_HIGH_TAG 4

#define MPI_MAX_COMMS 500
extern struct MPI_Comm_entry MPI_Comms[];

extern MPI_Comm MPI_first_available_comm;
extern MPI_Comm MPI_COMM_PARENT;

struct MPI_init_msg { /* Initial message from master to slave */
  MPI__INT len;       /* For consistency check;  This should equal sizeof(init_msg) */
  MPI__INT node_num;  /* here is node number == rank in MPI_COMM_WORLD */
  MPI__INT num_nodes;
};

/* Initial message from spwaning process to spawned processes */
struct MPI_init_msg1 { 
  MPI__INT len;          /* For consistency check;  This should equal sizeof(init_msg) */
  MPI__INT rank;         /* rank in spawned group */
  MPI__INT num_nodes;    /* number of nodes in spawning group */
  MPI__INT spawned_nodes;    /* number of spwaned nodes */
  MPI__INT send_context; /* MPI_Comm of spawning group, needed for making intercomm */
};
