#include "mpi.h"
#include "mpiimpl.h"

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/socket.h>
#include <sys/stat.h> 

/*======================================================================*/
#define MAX_PORTS 100

struct port_entry {
    int used;
    int port;
    struct sockaddr_in address;
    char port_name[MPI_MAX_PORT_NAME];
    int sd;
};

struct port_entry ports[MAX_PORTS];

/*======================================================================*/

void MPI_initPorts() {
  int i;
  for (i = 0; i < MAX_PORTS; i++)
    ports[i].used = 0;
}

/*======================================================================*/
	
int MPI_Open_port(MPI_Info info, char *port_name) {
  int i = 0;
  int len = MPI_processor_name_len;
  int sd;                    /* socket descriptor */
  char new_port[8];             /* port on which master listens, ASCII number */
  int port;
  struct sockaddr_in sin;      /* inet info structure            */

  while (ports[i].used)  
	i++;
  if (i == MAX_PORTS) {
    strcpy(port_name,"");
    return MPI_ERR_PORT;
  }

  MPI_new_listener(&sd, &sin, &port);
  sprintf(new_port, "%d", port);
  
  ports[i].port = port;
  ports[i].address = sin;
  ports[i].sd = sd;
  strcpy(ports[i].port_name, MPI_processor_name);
  strcpy(&(ports[i].port_name[len]), ":");
  strcpy(&(ports[i].port_name[len+1]), new_port);

  ports[i].used = 1;

  strcpy(port_name, ports[i].port_name);
  return MPI_SUCCESS;
}

/*======================================================================*/

static int findPortEntry(char *port_name) {
   int i;

   for (i = 0; i < MAX_PORTS; i++) {
     if (!ports[i].used)
       continue;
     if (!strcmp(port_name, ports[i].port_name)) {
        return i;
     }
  }
  return -1;
}
/*======================================================================*/

int MPI_Close_port(char *port_name) {
  int i = findPortEntry(port_name);
  if (i < 0 )
    return MPI_ERR_PORT;

  /* do not accept any more connections at this port */
  close(ports[i].sd);
  ports[i].used = 0;
  return MPI_SUCCESS;
}

/*======================================================================*/

int MPI_Comm_accept(char *port_name, MPI_Info info, int root,
			MPI_Comm comm, MPI_Comm *newcomm) {
  int i, d;
  struct MPI_Comm_entry local_entry, peer_entry, *new_entry;
  MPI_Comm remote_comm;
  MPI_Group remote_group = MPI_getGroup();
  int high;
  int remote_size;
  MPI_Status status;

  struct sockaddr_in sin;
  size_t len = sizeof(struct sockaddr_in);

  if (MPI_first_available_comm >= MPI_MAX_COMMS) {
    printf("Can not allocate more communicators\n");
    exit(1);
  }
  if (comm == MPI_COMM_NULL) {
    printf("Error: creating intercommunicator from non communicator\n");
    exit(1);
  }
  if (root == MPI_ANY_SOURCE) {
    printf("Error: creating intercommunicator using wild-cards\n");
    exit(1);
  }
    
  local_entry = MPI_Comms[comm];
  if (!local_entry.intracomm) {
    printf("Error: creating intercommunicator from non intracommunicator\n");
    exit(1);
  }

  if (root == local_entry.source) { /* root */
     int ns, port = findPortEntry(port_name);
     MPI__INT group_size;
     int node, size;
     MPI_Group group = MPI_Comms[comm].group;
     MPI_Comm available;

     if (port < 0) {
       printf("Eroor: Unexisting port\n");
       exit(1);
     }

     group_size = group->size;
     /* agree with local members on new comm */
     for (i = 0; i < group_size; i++) {
        if (i == root)
            continue;
        MPI_internal_rcv(comm, i, MPI_ACCEPT, NULL, 0, MPI_BYTE, &available, &d, &d, &status);
        if (available > MPI_first_available_comm)
            MPI_first_available_comm = available;
     }

     CALL_CHK(ns = accept, (ports[port].sd, (struct sockaddr*)&sin, &len));

     /* receive size of remote group */
     CALL_CHK(size = recv, (ns, (char *)&remote_size, sizeof(remote_size),
0));
     if (size != sizeof(remote_size)) {
	printf("Incomplete group size received\n");
	exit(1);
     }
     remote_size = ntohl(remote_size);
  
     if (MPI_num_nodes + remote_size >= MPI_MAX_NODES) {
       printf("Can not connect with that many nodes\n");
       exit(1);
     }  
     remote_group->size = remote_size;  

     /* receive remote comm */
     CALL_CHK(size = recv, (ns, (char *)&remote_comm, sizeof(remote_comm),
0));
     if (size != sizeof(remote_comm)) {
	printf("Incomplete message received\n");
	exit(1);
     }
     remote_comm = ntohl(remote_comm);

     /* receive addresses of remote group members */
     len = sizeof(struct sockaddr_in);
     for (i = 0; i < remote_size; i++) {
       CALL_CHK(size = recv, (ns, (char *)&MPI_nodes[MPI_num_nodes].listener_addr, len, 0));
       if (size != len) {
         printf("Incomplete address message received.\n");
         exit(1);
       }
       remote_group->nodes[i] = MPI_num_nodes;
       remote_group->nodes[MPI_num_nodes] = i;
       MPI_num_nodes++;
     }

     /* send size of local group */
     group_size = htonl(group->size);
     CALL_CHK(send, (ns, (char*)&group_size, sizeof(group_size), 0));

     /* send new communicator entry */
     {  MPI__INT comm = htonl(MPI_first_available_comm);
        CALL_CHK(send, (ns, (char*)&comm, sizeof(comm), 0)); 
     }

     /* send addresses of local group members */
     group_size = group->size;
     for (i = 0; i < group_size; i++) {
       node = group->nodes[i];
       CALL_CHK(send, (ns, (char *)&MPI_nodes[node].listener_addr, len, 0));
     }

     { /* Acknowledge that ... are synchronized */
       MPI__INT buf = htonl(len);
       CALL_CHK(send, (ns, (char *)&buf, sizeof(buf), 0));
     }
  }
  else { /* non-root */
    MPI_internal_snd(comm, root, MPI_ACCEPT, NULL, 0, MPI_BYTE, 
		MPI_first_available_comm, d, d);

    MPI_internal_rcv(comm, root, MPI_INTERCOMM_CREATE, NULL, 0, MPI_BYTE, 
		&MPI_first_available_comm, &high, &d, &status);

    MPI_internal_rcv(comm, root, MPI_INTERCOMM_CREATE, NULL, 0, MPI_BYTE, 
		&remote_comm, &(remote_group->size), &d, &status);

     for (i = 0; i < remote_group->size; i++) {
	/* put new remote group at the end */
	MPI_internal_rcv(comm, root, MPI_INTERCOMM_CREATE,	
	   (char *)&MPI_nodes[MPI_num_nodes].listener_addr, len,
	   MPI_BYTE, &d, &d, &d, &status);
       remote_group->nodes[i] = MPI_num_nodes;
       remote_group->nodes[MPI_num_nodes] = i;
       MPI_num_nodes++;
    }
  } /* DONE */

  *newcomm = MPI_first_available_comm++;
  new_entry = &MPI_Comms[*newcomm];
  
  new_entry->group = remote_group;
  new_entry->group->references++;
  MPI_Group_free(&remote_group);
  
  new_entry->local = local_entry.group;
  new_entry->local->references++;
  
  new_entry->rcv_context = *newcomm;
  new_entry->snd_context = remote_comm;

  new_entry->source = local_entry.source; /* keep the rank */
  new_entry->intracomm = 0;
  new_entry->used = 1;
  new_entry->tag = high ? MPI_HIGH_TAG : 0;

  return MPI_SUCCESS;
}

/*======================================================================*/
static void extractHostAndPort(char *port_name, char *host_name, int *port);

int MPI_Comm_connect(char *port_name, MPI_Info info, int root,
			MPI_Comm comm, MPI_Comm *newcomm) {

  int i, d;
  struct MPI_Comm_entry local_entry, peer_entry, *new_entry;
  MPI_Comm remote_comm;
  MPI_Group remote_group = MPI_getGroup();
  int high;
  int remote_size;
  MPI_Status status;

  struct sockaddr_in sin;
  size_t len = sizeof(struct sockaddr_in);

  *newcomm = MPI_COMM_NULL;

  if (MPI_first_available_comm >= MPI_MAX_COMMS) {
    printf("Can not allocate more communicators\n");
    exit(1);
  }
  if (comm == MPI_COMM_NULL) {
    printf("Error: creating intercommunicator from non communicator\n");
    exit(1);
  }
  if (root == MPI_ANY_SOURCE) {
    printf("Error: creating intercommunicator using wild-cards\n");
    exit(1);
  }
    
  local_entry = MPI_Comms[comm];
  if (!local_entry.intracomm) {
    printf("Error: creating intercommunicator from non intracommunicator\n");
    exit(1);
  }

  if (root == local_entry.source) { /* root */
     int ns;
     MPI__INT group_size;
     int node, size;
     MPI_Group group = MPI_Comms[comm].group;
     MPI_Comm available;
     struct hostent *hp;       /* ptr to host info structure*/

     char host[MPI_MAX_PROCESSOR_NAME];
     int port;
     extractHostAndPort(port_name, host, &port);

     group_size = group->size;
     /* agree with local members on new comm */
     for (i = 0; i < group_size; i++) {
        if (i == root)
            continue;
        MPI_internal_rcv(comm, i, MPI_ACCEPT, NULL, 0, MPI_BYTE, &available, &d, &d, &status);
        if (available > MPI_first_available_comm)
            MPI_first_available_comm = available;
     }


     remote_group->size = 0;
     /* get the host info such as IP address */
     if ((hp = gethostbyname(host)) == NULL) {
        MPI_propagateGroup(comm, root, remote_group, MPI_UNDEFINED, 0);
        return MPI_ERR_PORT;
     }

     bzero((char*)&sin, len);
     memcpy((char *)&sin.sin_addr, hp->h_addr, hp->h_length);
     sin.sin_port = htons(port); 
     sin.sin_family = hp->h_addrtype;

     /* we are using the default protocol 6 = IPPROTO_TCP */
     CALL_CHK(ns = socket, (AF_INET, SOCK_STREAM, IPPROTO_TCP));
     /* Connect to listener */
     if (0 != connect(ns, (struct sockaddr*)&sin, len)) {
        MPI_propagateGroup(comm, root, remote_group, MPI_UNDEFINED, 0);
        return MPI_ERR_PORT;
     }

     /* send size of local group */
     group_size = htonl(group->size);
     CALL_CHK(send, (ns, (char*)&group_size, sizeof(group_size), 0));

     /* send new communicator entry */
     {  MPI__INT comm = htonl(MPI_first_available_comm);
        CALL_CHK(send, (ns, (char*)&comm, sizeof(comm), 0)); 
     }

     /* send addresses of local group members */
     group_size = group->size;
     for (i = 0; i < group_size; i++) {
       node = group->nodes[i];
       CALL_CHK(send, (ns, (char *)&MPI_nodes[node].listener_addr, len, 0));
     }

     /* receive size of remote group */
     CALL_CHK(size = recv, (ns, (char*)&remote_size, sizeof(remote_size),
0));
     if (size != sizeof(remote_size)) {
	printf("Incomplete group size received\n");
	exit(1);
     }
     remote_size = ntohl(remote_size);  
     if (MPI_num_nodes + remote_size >= MPI_MAX_NODES) {
       printf("Can not connect with that many nodes\n");
       exit(1);
     }  
     remote_group->size = remote_size;  

     /* receive remote comm */
     CALL_CHK(size = recv, (ns, (char*)&remote_comm, sizeof(remote_comm),
0));
     if (size != sizeof(remote_comm)) {
	printf("Incomplete message received\n");
	exit(1);
     }
     remote_comm = ntohl(remote_comm);

     /* receive addresses of remote group members */
     len = sizeof(struct sockaddr_in);
     for (i = 0; i < remote_size; i++) {
       CALL_CHK(size = recv, (ns, (char *)&MPI_nodes[MPI_num_nodes].listener_addr, len, 0));
       if (size != len) {
         printf("Incomplete address message received.\n");
         exit(1);
       }
       remote_group->nodes[i] = MPI_num_nodes;
       remote_group->nodes[MPI_num_nodes] = i;
       MPI_num_nodes++;
     }

     {  MPI__INT buf;
        CALL_CHK(recv, (ns, (char *)&buf, sizeof(buf), 0));
        if (ntohl(buf) != len) {
          printf("MPI_Connect: connection %d not synchronized.\n", MPI_node_num);
          exit(1);
        }
     }

     /* send to the local members remote group */
     MPI_propagateGroup(comm, root, remote_group, remote_comm, (high = 0));
  }
  else { /* non-root */
    MPI_Comm available;
    MPI_internal_snd(comm, root, MPI_ACCEPT, NULL, 0, MPI_BYTE, 
		MPI_first_available_comm, d, d);

    MPI_internal_rcv(comm, root, MPI_INTERCOMM_CREATE, NULL, 0, MPI_BYTE, 
		&available, &high, &d, &status);

    MPI_internal_rcv(comm, root, MPI_INTERCOMM_CREATE, NULL, 0, MPI_BYTE, 
		&remote_comm, &(remote_group->size), &d, &status);

    if (remote_comm == MPI_UNDEFINED) {
	/* something went wrong, but not fatal at root */
        return MPI_ERR_PORT;
    }
    available = MPI_first_available_comm;

     for (i = 0; i < remote_group->size; i++) {
	/* put new remote group at the end */
	MPI_internal_rcv(comm, root, MPI_INTERCOMM_CREATE,	
	   (char *)&MPI_nodes[MPI_num_nodes].listener_addr, len,
	   MPI_BYTE, &d, &d, &d, &status);
       remote_group->nodes[i] = MPI_num_nodes;
       remote_group->nodes[MPI_num_nodes] = i;
       MPI_num_nodes++;
    }
  } /* DONE */

  *newcomm = MPI_first_available_comm++;
  new_entry = &MPI_Comms[*newcomm];
  
  new_entry->group = remote_group;
  new_entry->group->references++;
  MPI_Group_free(&remote_group);
  
  new_entry->local = local_entry.group;
  new_entry->local->references++;
  
  new_entry->rcv_context = *newcomm;
  new_entry->snd_context = remote_comm;

  new_entry->source = local_entry.source; /* keep the rank */
  new_entry->intracomm = 0;
  new_entry->used = 1;
  new_entry->tag = high ? MPI_HIGH_TAG : 0;

  return MPI_SUCCESS;
}

/*========================================================================*/
static void extractHostAndPort(char *port_name, char *host_name, int *port) {
   int pos;
   char *tmp;

   tmp = strchr(port_name, ':');
   if (tmp == NULL) {
     *port = -1;
     strcpy(host_name, "");
     return;
   }
   pos = tmp - port_name;
   strncpy(host_name, port_name, pos);
   host_name[pos] = '\000';
   *port = atoi(&port_name[pos+1]);  
}

/*========================================================================*/
