#include "mpi.h"
#include "mpiimpl.h"

#define __USE_POSIX
#include <signal.h>
/*
#include <sigset.h>
*/
#include <sys/time.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <stdio.h>

/* NOTE: I get "`sigset_t' undeclared" message if I turn the ANSII
   switch on compiler
*/

#define unDEBUG
/*========================================================================================*/
typedef void (*sighandler_t)(int);

static sig_atomic_t IsSIGINT;

static void null_handler(int signum) {
  IsSIGINT = 1;
}

/* Under POSIX, the system calls used here that are interruptible (EINTR)
   are:  connect), recv(), select().
   For them we can use CALL_CHK.

   GNU C library manual says :
     Any signal will cause select to return immediately. 
     So if your program uses signals, you can't rely
     on select to keep waiting for the full time specified.
     If you want to be sure of waiting for a particular
     amount of time, you must check for EINTR and repeat the select
     with a newly calculated timeout based on the current time.
  We are conserned about select returning to us only partial result 
  when some signal rise.
*/
static int safe_select(int max, fd_set *readfds, fd_set *writefds, 
               fd_set *exceptfds, struct timeval *timeout ) {
  sigset_t sigset, oldsigset;
  sighandler_t real_handler;
  int answer;
  
  do {
    sigfillset(&sigset);
    sigdelset(&sigset, SIGINT);
    /* SO_KEEPALIVE can generate SIGPIPE.  If this fnc is called
       through mpiimpl.h:CALL_CHK() let it re-start this automatically.
       But call to select should not generate SIGPIPE !
    */
    sigdelset(&sigset, SIGPIPE);

    real_handler = signal(SIGINT, null_handler);
    sigprocmask(SIG_BLOCK, &sigset, &oldsigset);
    
    IsSIGINT = 0;

#ifdef _DEBUG
    printf("Start select max fd %d\n", max);
#endif
    answer = select(max, readfds, writefds, exceptfds, timeout);

    sigprocmask(SIG_SETMASK, &oldsigset, NULL);
    signal(SIGINT, real_handler);

    if (IsSIGINT) 
      raise(SIGINT); /* let the user handle it */
  } while (IsSIGINT == 1); /* null_handler could have set IsSIGINT = 1 */

#ifdef _DEBUG
  printf("Select return %d and IsSIGINT %d\n", answer, IsSIGINT);
  printf("Errno %d , %d %d %d \n", errno, EBADF, EINTR, EINVAL);
#endif
  return answer;
}

/*========================================================================================*/
static void accept_new_sender(int list_sd) {
  static void check_for_intruders(struct sockaddr_in);

  struct sockaddr_in new_sin;
  int new_sd;
  int fromlen = sizeof(new_sin);

  MPI__INT buf;
  MPI_Comm real_context, context;
  int source;
  int source_node; /* client's node number */
  struct MPI_node_struct *src;

#ifdef _DEBUG
  printf("Node %d accepting new socket\n", MPI_node_num); 
#endif

  CALL_CHK(new_sd = accept, (list_sd, (struct sockaddr *)&new_sin, &fromlen));

  check_for_intruders(new_sin);

  CALL_CHK(recv, (new_sd, (char *)&buf, sizeof(buf), 0));
  real_context = ntohl(buf);
  context = abs(real_context);

  if (!MPI_Comms[context].used) {
    printf("accept_new_client: Bad comm number received: %d\n", context);
    exit(1);
  }

  CALL_CHK(recv, (new_sd, (char *)&buf, sizeof(buf), 0));
  source = ntohl(buf); /* source rank */

  if (real_context >= 0)
     source_node = MPI_Comms[context].group->nodes[source];
  else
     source_node = MPI_Comms[context].local->nodes[source];

  if (source_node == MPI_UNDEFINED) {
    printf("accept_new_client: Bad source %d in comm :%d\n", source, context);
    exit(1);
  }
  
#ifdef DEBUG
  printf("Node %d accepted socket %d from node %d\n", MPI_node_num, new_sd, source_node); 
#endif

  src = &MPI_nodes[source_node];

  if (src->sd != MPI_NOSOCKET) {
    FD_CLR(src->sd, &MPI_fdset);
    if (MPI_is_connected(src->sd))
      /* i should read all messages in local receive buffer */
      CALL_CHK(shutdown, (src->sd, 0));
  }
  src->sd = new_sd;
  MPI_new_sd(new_sd);
}

/*========================================================================================*/

static void check_for_intruders(struct sockaddr_in sin) {
}

/*========================================================================================*/

static void setFDset(MPI_Comm comm, int source, fd_set *fdset) {
  int fd, i, size, node;
  MPI_Group group;
  
  if (comm >= 0)
    group = MPI_Comms[comm].group;
  else
    group = MPI_Comms[-comm].local;

  FD_ZERO(fdset);

/* we need this always if we don't want to check if every socket is alive, 
that can be very time expensive for ANY_SOURCE. In this way, we will maybe 
accept som connections out of "our" order, but no extra job will be done.
We eould have to accept those connections eventualy. *
  FD_SET(MPI_my_list_sd, fdset);
*/

  if (source != MPI_ANY_SOURCE) {
    fd = MPI_nodes[group->nodes[source]].sd;
/*
    FD_SET(((fd == MPI_NOSOCKET) ? MPI_my_list_sd : fd), fdset);
*/
    if ((fd == MPI_NOSOCKET) || !MPI_is_connected(fd))
	FD_SET(MPI_my_list_sd, fdset);
    else  
	FD_SET(fd, fdset);

    return;
  }
  
  size = group->size;
  for (i = 0; i < size; i++) {
    node = group->nodes[i];
    if (node == MPI_node_num)
      continue;
    fd = MPI_nodes[node].sd;
/*
    FD_SET(((fd == MPI_NOSOCKET) ? MPI_my_list_sd : fd), fdset);
*/
    if ((fd == MPI_NOSOCKET) || !MPI_is_connected(fd))
	FD_SET(MPI_my_list_sd, fdset);
    else  
	FD_SET(fd, fdset);
  }
}

/*========================================================================================*/

static void bad_hdr() {
  printf("Bad header\n");
}

/*========================================================================================*/

int MPI_read_msg_hdr(int source, int tag, MPI_Comm comm, 
             int *flag, struct MPI_msg_hdr *msg_hdr, int *fd,
             int blocking_flag, int eat_hdr_flag) {

  size_t hdr_size = sizeof(*msg_hdr);
  int recv_flag, num_bytes, ready4read;
  fd_set curr_fdset;
  struct timeval zerotime, *timeout;
  
#ifdef _DEBUG
  if (tag >= 0)
    printf("Peeking in the msg header comm = %d, source = %d, "
       "tag = %d, at node %d (DONT_EAT_HDR=%d)\n",
       comm, source, tag, MPI_node_num, eat_hdr_flag);
#endif

  /* check buffer first*/
  if (MPI_peek_rcv_buff(comm, tag, source, msg_hdr) == 0) {
    *flag = MPI_MSG_WAITING; /* message is already buffered */
#ifdef DEBUG
    printf("msg waiting in buffer\n");
#endif
    return 0; /* success */
  }
#ifdef _DEBUG
  printf("Not found in buffer\n");
#endif
  
  if (blocking_flag == MPI_BLOCKING)
    timeout = NULL; /* wait forever (block) */
  else {            /* poll and return */
    zerotime.tv_sec = 0;
    zerotime.tv_usec = 0;
    timeout = &zerotime;
  }

  do {
    /* Use select to find active port; if BLOCKING && not MPI_ANY_SOURCE,
       then we could have skipped this and gone directly to recv(), except
       that we also need to check the listener port */
    setFDset(comm, source, &curr_fdset);
    ready4read = safe_select(MPI_max_sd + 1, &curr_fdset, NULL, NULL, timeout);

#ifdef DEBUG
    printf("%d socket(s) ready for read at node %d\n",ready4read, MPI_node_num);
#endif
  
     if ((ready4read == 0) && (*flag == MPI_BLOCKING)) {/* this realy should not happen */
      printf("read_msg_hdr: BLOCKING exiting without recv'd msg.\n");
      exit(1);
    }

    if (ready4read == 0) { /* && (*flag != MPI_BLOCKING) */
      *flag = MPI_NO_MSG_WAITING;  /* no messages, return */
      return (-1);
    }

    /* A msg on this socket is a request to open a new socket. */
    if (FD_ISSET(MPI_my_list_sd, &curr_fdset)) {
#ifdef DEBUG
      printf("Listener ready to connect at socket %d at node %d\n", 
         MPI_my_list_sd, MPI_node_num); 
#endif      
      accept_new_sender(MPI_my_list_sd);
      continue; /* poll again */
    }
    
    /* At this point curr_fdset containes all sockets that match source
       and are ready for reading */    
    recv_flag = (eat_hdr_flag == MPI_DONT_EAT_HDR) ? MSG_PEEK : 0;

    *fd = 0;
    do {
      /* find the socket to be checked now */
      if (source == MPI_ANY_SOURCE){ /* Determine source, fd */
        for (; *fd <= MPI_max_sd + 1; (*fd)++)
          if (FD_ISSET(*fd, &curr_fdset)) { 
		/* this socket is ready for reading */
#ifdef _DEBUG
        printf("Socket %d ready for reading at node %d\n", *fd, MPI_node_num);
#endif
              break; /* break from "for" loop */
          }
      }
      else /* concrete source */ {
	if (comm >= 0)
    	  *fd = MPI_nodes[MPI_Comms[comm].group->nodes[source]].sd;
	else
    	  *fd = MPI_nodes[MPI_Comms[-comm].local->nodes[source]].sd;
      }
#ifdef DEBUG
      printf("Message waiting at node %d at socket %d max_sd %d\n",MPI_node_num, *fd, MPI_max_sd);
#endif      

      if (*fd > MPI_max_sd) { /* test if for loop has finished */
    printf("Unknown socket descriptor from select #%d max_sd = %d source = %d.\n", 
           *fd, MPI_max_sd, source);
      exit(1);
      }

      CALL_CHK(num_bytes = recv, (*fd, (char *)msg_hdr, hdr_size,
recv_flag));

#ifdef _DEBUG
      printf("Message at node %d from socket %d length %d\n", MPI_node_num, *fd, num_bytes);
#endif      

      /* If 0 characters received and another socket available */
      if (num_bytes == 0 && ready4read > 1 ) {
    FD_CLR(*fd, &curr_fdset);
    ready4read--;
#ifdef DEBUG
    printf("false alarm, test next socket\n");
#endif
    continue; /* false alarm, test next socket */
      }
      
      /* if no msg, non-blocking and no-one else ready for reading */
      if (num_bytes == 0 && timeout != NULL) { /* && ready4read == 1 */
	*flag = MPI_NO_MSG_WAITING;  /* no messages, return */
    	return (-1);
      }
      /* if 0 characters received, blocking call, and no-one else ready for reading */
      if (num_bytes == 0 && timeout == NULL) {/* && ready4read == 1 */  
#ifdef DEBUG
    printf("Close socket! 0 characters received, blocking call, and no-one else ready for reading \n");
#endif
    	shutdown(*fd, 2);
    	break; /* poll again */
      }

      if (num_bytes != hdr_size) { /* did not received complete msg header */
    	bad_hdr(); /* Possible non-local error or loss of socket? */
    	exit (1);
      }
	/* translate message header to local format */
      msg_hdr->comm = ntohl(msg_hdr->comm);
      msg_hdr->source = ntohl(msg_hdr->source);
      msg_hdr->tag = ntohl(msg_hdr->tag);
      msg_hdr->datatype = ntohl(msg_hdr->datatype);
      msg_hdr->size = ntohl(msg_hdr->size);
      msg_hdr->info1 = ntohl(msg_hdr->info1);
      msg_hdr->info2 = ntohl(msg_hdr->info2);
      msg_hdr->info3 = ntohl(msg_hdr->info3);

      if (msg_hdr->size == -1) { 
	printf("size of message was -1\n"); 
	exit(1); 
      }

      /* If message is received from different communicator or with incompatible requested tag 
	or message type put it in the receive buffer */
      if ((msg_hdr->comm != abs(comm)) || 
      ((tag != MPI_ANY_TAG) && (tag != msg_hdr->tag))) {
    void *buf = MPI_put_rcv_buff(msg_hdr);
    int size;
    if (buf == NULL) {
      printf("Msg received with incompatible requested tag, "
         "and not enough space in the receive buffer\n");
      exit(1);
    }
    size = msg_hdr->size;
    if (recv_flag == MSG_PEEK) { /* retrive the buffer */
      CALL_CHK(num_bytes = recv,(*fd, (char *)msg_hdr, hdr_size, 0));
      if (num_bytes != hdr_size) {
        bad_hdr();
        exit (1);
      }
    }
    while (size > 0) {
      CALL_CHK(num_bytes = recv,(*fd, (char *)buf, size, 0));
      size -= num_bytes;
      buf = buf + num_bytes;
    }
#ifdef DEBUG
     printf("Buffered comm = %d tag = %d len = %d wanted comm = %d tag = %d len = %d \n",
          msg_hdr->comm, msg_hdr->tag, msg_hdr->size, comm, tag, size);
#endif
    break; /* find next socket ready for reading, possible this one again */
      }
      else {
    /* everything matched */
    *flag = MPI_MSG_WAITING;
    return 0;
      }      
    } while(1);
  } while (1);
  return -1;
}

/*========================================================================================*/

int MPI_internal_rcv(MPI_Comm comm, int source, int tag, void* buf, int size,
             MPI_Datatype datatype,
             MPI__INT *info1, MPI__INT *info2,MPI__INT *info3,
             MPI_Status *status) {

  struct MPI_msg_hdr msg_hdr;
  void *tmp;
  size_t hdr_size = sizeof(struct MPI_msg_hdr);
  int flag, fd;

#ifdef DEBUG
  printf("Receiving msg with: comm = %d, source = %d, "
     "tag = %d, type = und size = %d at node %d",
     comm, source, tag, size, MPI_node_num);
  if (source != MPI_ANY_SOURCE)
    printf(" from node %d\n", MPI_Comms[comm].group->nodes[source]);
  else
    printf("\n");
#endif

  if (MPI_peek_rcv_buff(comm, tag, source, &msg_hdr) == 0) {
    /* message is already in receive buffer */
    if (MPI_get_rcv_buff(comm, tag, source,
             size, buf, &msg_hdr) == MPI_SUCCESS) {
      status->MPI_SOURCE = msg_hdr.source;
      status->MPI_TAG = msg_hdr.tag;
      status->mpi_size = msg_hdr.size;
      *info1 = msg_hdr.info1;
      *info2 = msg_hdr.info2;
      *info3 = msg_hdr.info3;
#ifdef _DEBUG
      printf("Message was in the buffer\n");
#endif
      return MPI_SUCCESS;
    }
    else {
#ifdef DEBUG
      printf("Message first found with a peek in buffer, but later not!\n");
#endif
    }
  }
  
  /* do regular stuff here */
  if (MPI_read_msg_hdr(source, tag, comm, &flag, &msg_hdr, &fd,
               MPI_BLOCKING, MPI_EAT_HDR) == 0 ){
    int i; 
    int msg_size = (msg_hdr.size > size ) ? size : msg_hdr.size;
    
    while (msg_size > 0) {
      CALL_CHK(i = recv, (fd, (char *)buf, msg_size, 0));
      buf = (char *)buf + i;
      msg_size -= i;
    }
    status->MPI_TAG = msg_hdr.tag;
    status->MPI_SOURCE = msg_hdr.source;
    status->mpi_size = msg_hdr.size;
    *info1 = msg_hdr.info1;
    *info2 = msg_hdr.info2;
    *info3 = msg_hdr.info3;
#ifdef _DEBUG
    printf("Message received\n");
#endif

#if 0 /* For debugging, kill a socket and see if MPI recovers */
    { 
      static int cont = 1;
      int r;
      if (cont) {
         printf("socket %d connected?: %d\n", fd, MPI_is_connected(fd));
         printf("KILLING socket %d from %d to %d for debugging\n",
           fd, status->MPI_SOURCE, MPI_node_num);
         r = shutdown(fd, 2); /* FOR DEBUGGING */
         printf("socket %d connected?: %d, %d\n", fd, MPI_is_connected(fd), r);
         cont = 0;
      }
      else
	printf("received msg. at socket %d\n", fd);
    }
#endif
    return MPI_SUCCESS;    
  }
  printf("Could not receive the message\n");
  exit (1);
}
