#include "mpi.h"
#include "mpiimpl.h"

#define __USE_POSIX
#include <signal.h>
/*
#include <sigset.h>
*/
#include <sys/time.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <stdio.h>

/* NOTE: I get "`sigset_t' undeclared" message if I turn the ANSII
   switch on compiler
*/

/*========================================================================================*/

static void connect_to_socket(MPI_Comm comm, int dest_node_num) {
  int sd;
  MPI__INT buf;
  struct MPI_node_struct *dest = &MPI_nodes[dest_node_num];
  size_t size = sizeof(struct sockaddr_in);

#ifdef DEBUG
  printf("Node %d connecting new socket for %d\n", MPI_node_num, dest_node_num);
#endif

  CALL_CHK(sd = socket, (AF_INET, SOCK_STREAM, IPPROTO_TCP));

  CALL_CHK(connect, (sd, (struct sockaddr *)&(dest->listener_addr), size));

  size = sizeof(buf);
  if (comm >= 0)
     buf = htonl(MPI_Comms[comm].snd_context);
  else
     buf = htonl(comm);

  CALL_CHK(send, (sd, (char *)&buf, size, 0));

  buf = htonl(MPI_Comms[abs(comm)].source);
  CALL_CHK(send, (sd, (char *)&buf, size, 0));

#ifdef DEBUG
  printf("Node %d connected new socket %d for %d\n", MPI_node_num, sd, dest_node_num);
#endif

  if (dest->sd != MPI_NOSOCKET ) {
    FD_CLR(dest->sd, &MPI_fdset );
    if(MPI_is_connected(dest->sd))
       CALL_CHK(shutdown, (dest->sd, 0));
  }
  dest->sd = sd;
  MPI_new_sd(sd);
}

/*========================================================================================*/
/*   send() to disconnected socket can generate SIGPIPE signal */

static int send_to;             /* number of process to whome msg is send */
static MPI_Comm context;		/* in wich context */

static sig_atomic_t isSIGPIPE; 

#define DEBUG

static void pipe_handler(int signum) {
#ifdef DEBUG
    printf("At node %d socket broken for process %d\n", MPI_node_num, send_to);
#endif
    connect_to_socket(context, send_to);
    isSIGPIPE = 1;
#ifdef DEBUG
    printf("New socket is %d\n", MPI_nodes[send_to].sd);
#endif
}

#undef DEBUG
/*========================================================================================*/
typedef void (*sighandler_t)(int);

/* this function should do the safe send, by catching SIGPIPE signal
   and opening the socket again */

static int safe_send(MPI_Comm comm, int to_process, 
			char *msg, int len, int flags) {
  sigset_t sigset, oldsigset;
  sighandler_t real_handler;
  int sd, answer;  

  if (len == 0)
    return;

  sd = MPI_nodes[to_process].sd;
  if (sd == MPI_NOSOCKET) {/* never used before */
#ifdef DEBUG  
    printf("Sending first msg from %d to %d\n", MPI_node_num, to_process);
#endif
    connect_to_socket(comm, to_process);
  }
#ifdef _DEBUG  
  sd = MPI_nodes[to_process].sd;
  printf("Send message at socket %d to node %d from node %d\n", sd, to_process, MPI_node_num);
#endif
  /* set signal */
  sigfillset(&sigset);
  sigdelset(&sigset, SIGPIPE);

  /* pipe_handler should re-establish connection if SIGPIPE is generated by send */
  send_to = to_process;
  context = comm;
  real_handler = signal(SIGPIPE, pipe_handler);

  sigprocmask(SIG_BLOCK, &sigset, &oldsigset);

  do {
    sd = MPI_nodes[to_process].sd; /* sd can be changed by signal handler */ 
    isSIGPIPE = 0;
    answer = send(sd, msg, len, flags);
  } while (isSIGPIPE == 1);
  
  isSIGPIPE = 0;
  /* revert signal */
  sigprocmask(SIG_SETMASK, &oldsigset, NULL);
  signal(SIGINT, real_handler);

  return answer;
}

/*========================================================================================*/

int MPI_internal_snd(MPI_Comm comm, int dest, int tag, 
		void* buf, int size,
             MPI_Datatype datatype, /*MPI__INT msg_type,*/
             MPI__INT info1, MPI__INT info2, MPI__INT info3) {
  void *tmp;
  struct MPI_msg_hdr hdr;
  size_t hdr_size = sizeof(struct MPI_msg_hdr);

  int target, real_comm = comm;

  if (comm >= 0)
    target = MPI_Comms[comm].group->nodes[dest];
  else {
    comm = -comm;
    target = MPI_Comms[comm].local->nodes[dest];
  }

if (comm < 0)
  exit(987);

  if (real_comm >= 0)
    hdr.comm = htonl(MPI_Comms[comm].snd_context); /* remote comm number */
  else
    hdr.comm = htonl(comm); /* local comm number */

  hdr.source = htonl(MPI_Comms[comm].source);    /* my rank in my local group */
  hdr.tag = htonl(tag);
  hdr.datatype = htonl(datatype);
  hdr.size = htonl(size);     /* size is in bytes */
  hdr.info1 = htonl(info1);    /* additional fields for special use */
  hdr.info2 = htonl(info2);
  hdr.info3 = htonl(info3);
  
#ifdef DEBUG
  printf("Sending msg with: comm = %d, dest = %d, dest_node = %d,"
     " tag = %d, size = %d at node %d\n",
     real_comm, dest, target, tag, size, MPI_node_num);
#endif

  if (target == MPI_node_num) { /* sending to myself */
    /* return header to local format */
    hdr.comm = MPI_Comms[comm].snd_context;
    hdr.source = MPI_Comms[comm].source;
    hdr.tag = tag;
    hdr.datatype = datatype;
    hdr.size = size;
    hdr.info1 = info1;
    hdr.info2 = info2;
    hdr.info3 = info3;
    tmp = MPI_put_rcv_buff(&hdr);
    if (tmp == NULL) {/* no more room in buffer */
      printf("Message sent to myself can not fit into the buffer %d\n", MPI_node_num);
      exit(1); /* would deadlock */
    }
    memcpy(tmp, buf, size);
    return MPI_SUCCESS;
  }

  safe_send(real_comm, target, (char *)&hdr, hdr_size, 0);
  safe_send(real_comm, target, (char *)buf, size, 0);

#ifdef _DEBUG
  printf("Sent msg with: comm = %d, dest_node = %d,"
     " tag = %d, size = %d at node %d\n",
     real_comm,  target, tag, size, MPI_node_num);
#endif

#if 0 /* For debugging, kill a socket and see if MPI recovers */
    { 
      static int cont = 1;
      int fd = MPI_nodes[target].sd;
      int r;
      if (cont) { /* && (MPI_node_num==0)) {*/
         printf("socket %d connected?: %d\n", fd, MPI_is_connected(fd));
         printf("KILLING socket %d from %d to %d for debugging\n",
           fd, MPI_node_num, target);
         r = shutdown(fd, 2); /* FOR DEBUGGING */
         printf("socket %d connected?: %d, %d\n", fd, MPI_is_connected(fd), r);
         cont = 0;
      }
      else
	printf("sent msg. to socket %d\n", fd);
    }
#endif
  return MPI_SUCCESS;
}
/*========================================================================================*/
