#include "mpi.h"
#include "mpiimpl.h"

#include <netinet/in.h>
#include <netdb.h>

#define unDEBUG

/* Spawned processes will be spawned at the root's host
   MPI_Info is ignored  */
int MPI_Comm_spawn(char *command, char *argv[], int maxprocs,
           MPI_Info info, int root, MPI_Comm comm, 
           MPI_Comm *intercomm, int array_of_errcodes[]) {
                
  static void spawn_children(char *command, char *argv[], int procs, MPI_Group group);
  struct MPI_Comm_entry entry, *intercomm_entry;
  MPI_Group group;
  MPI_Status status;
  int i, d, num_proc, tag = 0;

  struct sockaddr_in new_sin;
  size_t addr_size = sizeof(struct sockaddr_in);
  size_t fromlen; 
  
  if (comm == MPI_COMM_NULL) {
    printf("Error: spawn from non communicator\n");
    exit(1);
  }
  if (MPI_first_available_comm == MPI_MAX_COMMS) {
    printf("Error: Can not allocate more communicators\n");
    exit(1);
  }
  entry = MPI_Comms[comm];
  if (!entry.intracomm) { /* can spawn only from intracomms */
    printf("Error: spawn from intercommunicator\n");
    exit(1);
  }  

  if (entry.source == root) { /* root */
    int nodes, nodes2spawn;
    MPI_Comm available;
    int j, sd, newport;
    int group_size = entry.group->size;
    
#ifdef DEBUG
    printf("Root now entering spawn function \n");
    printf("Group size is %d\n",group_size );    
#endif
    /* find out max of first available comm entries and min # of available nodes */
    nodes2spawn = MPI_MAX_NODES - MPI_num_nodes;

    for (i = 0; i < group_size; i++) {
	if (i == root)
		continue;
      MPI_internal_rcv(comm, i, MPI_COMM_SPAWN, NULL, 0, MPI_BYTE,
                &available, &nodes, &d, &status);
      if (available > MPI_first_available_comm)
        MPI_first_available_comm = available;
      if (nodes < nodes2spawn)
        nodes2spawn = nodes;
    }
#ifdef DEBUG
    printf("Agread new Comm number is %d and # of av. nodes is %d\n", 
       MPI_first_available_comm, nodes2spawn);
#endif 
    
    num_proc = (maxprocs <= nodes2spawn) ? maxprocs : nodes2spawn;
    
    /* deal with spawned processes */
    spawn_children(command, argv, num_proc, entry.group);
    MPI_slaves += num_proc;

    for (i = 0; i < group_size; i++) {
      if (i == root)
	    continue;             
      /* send max available comm entry, and where process numbers of spawned 
         processes begin so they can create group for them */
      MPI_internal_snd(comm, i, MPI_COMM_SPAWN, NULL, 0, MPI_BYTE,
                MPI_first_available_comm, num_proc, d);
      /* send listener addresses of spawned children to the non-roots */
      for (j = 0; j < num_proc; j++)
    	MPI_internal_snd(comm, i, MPI_COMM_SPAWN, 
		(char *)&MPI_nodes[MPI_num_nodes + j].listener_addr, 
             	addr_size, MPI_BYTE, d, d, d);
    }
#ifdef DEBUG
    printf("Root has finished spawn function \n");
#endif 
  }
  else { /* non-root */
#ifdef DEBUG
    printf("Non-root now entering spawn function \n");
#endif 
 
    MPI_internal_snd(comm, root, MPI_COMM_SPAWN, NULL, 0, MPI_BYTE,
              MPI_first_available_comm, MPI_MAX_NODES - MPI_num_nodes, d);

    MPI_internal_rcv(comm, root, MPI_COMM_SPAWN, NULL, 0, MPI_BYTE, 
             &MPI_first_available_comm, &num_proc, &d, &status);
    
    /* receive listener addresses of spawned children from */
#ifdef DEBUG
    printf("Non root knows %d spawned processes\n", num_proc);
#endif
    for (i = 0; i < num_proc; i++) 
      MPI_internal_rcv(comm, root, MPI_COMM_SPAWN,
               (char *)&MPI_nodes[MPI_num_nodes + i].listener_addr, addr_size, 
               MPI_BYTE, &d, &d, &d, &status);
#ifdef DEBUG
    printf("Non-Root has finished spawn function \n");
#endif 
  }
  
  /* now make an intercomm for it */
  *intercomm = MPI_first_available_comm++;
  intercomm_entry = &MPI_Comms[*intercomm];
  
  group = MPI_getGroup();
  for (i = 0; i < num_proc; i++) {
    group->ranks[MPI_num_nodes + i] = i;
    group->nodes[i] = MPI_num_nodes + i;
  }
  group->size = num_proc;
    
  intercomm_entry->group = group;
  group->references ++;
  MPI_Group_free(&group);
  
  intercomm_entry->local = entry.group;
  intercomm_entry->local->references++;
  
  intercomm_entry->source = entry.source;
  intercomm_entry->rcv_context = *intercomm; 
  /* slaves create this after creating a MPI_COMM_WORLD */

  intercomm_entry->snd_context = 1; /* that is always parent entry in spawned processes 
                       0 is MPI_COMM_WORLD */
  intercomm_entry->intracomm = 0;
  intercomm_entry->used = 1;
  intercomm_entry->tag = MPI_HIGH_TAG;

  MPI_num_nodes += num_proc;
 
  if (array_of_errcodes != MPI_ERRCODES_IGNORE) {
    for (i = 0; i < num_proc; i++)
	array_of_errcodes[i] = MPI_SUCCESS;
    for (i = num_proc; i < maxprocs; i++)
	array_of_errcodes[i] = MPI_ERR_SPAWN;
  }
  return MPI_SUCCESS;  
}

/*========================================================================================*/

/* deal with spawned processes */
/* how to get connections from spawned but not to mix up with others? */
/* create new port !*/
static void spawn_children(char *command, char *argv[], int num_proc, MPI_Group group) {

  int argc;                             /* number of arguments */
  char tmp[MPI_PACK_GROUP_SIZE];        /* for packing group */
  char new_port[8]; /* port on which root listens, ASCII number */
  int i, ns, sd, port;
  struct hostent *hp;

  struct sockaddr_in new_sin;
  size_t addr_size = sizeof(struct sockaddr_in);
  size_t fromlen; 
  
  struct MPI_init_msg1 msg1;
  size_t msg1_size = sizeof(struct MPI_init_msg1);

  MPI_new_listener(&sd, &new_sin, &port);
  sprintf(new_port, "%d", port);
    
  argc = 0;
  if (argv != MPI_ARGV_NULL)
    while (argv[argc] != NULL) argc++;

  hp = gethostbyname(MPI_processor_name);

  for (i = 0; i < num_proc; i++) 
    MPI_fork_process(MPI_processor_name, command, argc, argv, 
             hp->h_name, new_port, SPAWNED_TAG);

  for (i = 0; i < num_proc; i++) { 
    fromlen = addr_size;
    CALL_CHK(ns = accept, (sd, (struct sockaddr *)&new_sin, &fromlen));
    
#ifdef DEBUG
    printf("Spawn root accepted socket %d for node %d\n", ns, MPI_num_nodes + i); 
#endif

    MPI_nodes[MPI_num_nodes + i].sd = ns;
    MPI_new_sd(ns);
    
    /* receive slave's listener_addr and store in field of pg_array[node] */
    CALL_CHK(recv, (ns, (char *)&(MPI_nodes[MPI_num_nodes +i].listener_addr), addr_size, 0));
    
    /* send slaves node number, rank, total number of nodes and spawning group */
    msg1.len = htonl(msg1_size);
    msg1.rank = htonl(i);
    msg1.spawned_nodes = htonl(group->size);
    msg1.num_nodes = htonl(num_proc);
    /* comm # children will use to communicate with parents */
    msg1.send_context = htonl(MPI_first_available_comm); 
    CALL_CHK(send, (ns, (char *)&msg1, msg1_size, 0));
#ifdef DEBUG
    printf("Spawn root has sent to node %d first message\n", MPI_num_nodes
+ i); 
#endif
  }
  
  { /* Send listener_addr fields of every node to slave */
    int i, j;    
    for (i = 0; i < num_proc; i++) {
      /* send adresses of spawned nodes first */
      for (j = 0; j < num_proc; j++)
         CALL_CHK(send, (MPI_nodes[MPI_num_nodes + i].sd, 
            	(char *)&MPI_nodes[MPI_num_nodes + j].listener_addr,
		addr_size, 0));

      /* send adresses of spawning nodes second */
      for (j = 0; j < group->size; j++ ) {
	int node = group->nodes[j];
         CALL_CHK(send, (MPI_nodes[MPI_num_nodes + i].sd, 
            	(char *)&MPI_nodes[node].listener_addr,
		addr_size, 0));
      }
    }
    { /* Acknowledge that master and slave are synchronized */
      MPI__INT buf;
      int i;
      buf =  htonl(addr_size);
      for (i = 0; i < num_proc; i++ ) {
           CALL_CHK(send, (MPI_nodes[MPI_num_nodes + i].sd, (char *)&buf,
		sizeof(buf), 0));
	      MPI_nodes[MPI_num_nodes + i].sd = MPI_NOSOCKET;
     }
    }
  }
}
